/******************************************************************************
 *
 * dirhive.c - Dump NT registry hives keys. Part of the DirTools project.
 *
 * $Revision: 1.4 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <inttypes.h>
#include <getopt.h>
#include <wchar.h>
#ifdef _WIN32
#include <windows.h>
#endif

#include "utility.h"
#include "sd_dump.h"
#include "nttime.h"
#include "dirhive.h"
#include "macros.h"
#include "dt_memory.h"
#include "dt_error.h"
#include "dt_locale.h"
#include "sysheader.h"
#ifdef _WIN32
#include "winsid.h"
#endif


#if SIZEOF_OFF_T == 8
#ifdef _WIN32
#define PRIuOff "I64u"
#define PRIxOff "I64x"
#else
#define PRIuOff "llu"
#define PRIxOff "llx"
#endif
#else
#define PRIuOff "u"
#define PRIxOff "x"
#endif

#define MAX_RECURSE_DEPTH 50
#define FIELDSHEADERS "Offset|Valid|KeyNode|ClassName|Last Write|# Subkeys|\
# Values|Owner|Group|# DACE|# SACE|[SID DACE|Rights|...][SID SACE|Audits|\
...]|\n"
char stdin_filename[] = "(stdin)";
FILE *FOutput = NULL;
FILE *FDbgOut = NULL;

int walkback (struct hive *, off_t, int *, int *, off_t);

int
get_block_size (struct hive *hdesc, off_t goffset, int32_t *seglen)
{
     int result = 0;
     
     *seglen = 0;

     if ((goffset + 4) <= hdesc->size)
          *seglen = DT_GETS32 (hdesc->buffer + goffset);
     else {
          ALARM (TA_PROG_ERR, "block size do not"
                 " fit in file at offset 0x%" PRIxOff, goffset);
          return INVALID_HEADER_BLOCK;
     }

     if (*seglen < 0)
          *seglen = -*seglen;
     else
          result |= UNUSED_HEADER_BLOCK;

     if (*seglen < 4) {
          ALARM (TA_PROG_ERR, "block size is too small (%" PRId32 ")"
                 " at offset 0x%" PRIxOff, *seglen, goffset);
          return INVALID_HEADER_BLOCK;
     }

     if ((goffset + *seglen) > hdesc->size) {
          ALARM (TA_PROG_ERR, "end of block beyond"
                 " end of file at offset 0x%" PRIxOff, goffset);
          return INVALID_HEADER_BLOCK;
     }

     return result;
}

int
check_block_header (struct hive *hdesc, off_t goffset,
                    int32_t * seglen, uint16_t * id, void **infos)
{

     int result = 0, bsize = 0;

     *id = 0;
     *infos = NULL;

     if ((result = get_block_size (hdesc, goffset, seglen)) ==
          INVALID_HEADER_BLOCK) {
	  return INVALID_HEADER_BLOCK;
     }

     if (((goffset + 6) > hdesc->size) || (*seglen < 6)) {
          ALARM (TA_PROG_ERR, "unable to read block id"
                 " at offset 0x%" PRIxOff, goffset);
          return INVALID_HEADER_BLOCK;
     }

     *id = DT_GETU16 (hdesc->buffer + goffset + 4);

     if (FDbgOut) {
          fprintf (FDbgOut, "check_block_header: offset 0x%" PRIxOff
                   ", size %" PRId32 ", %sused, id %" PRIx16 "\n", goffset,
                   *seglen, (result & UNUSED_HEADER_BLOCK ? "un" : ""), *id);
     }

     switch (*id) {
     case 0x6b6e:
          bsize = sizeof (struct nk_key);
          break;
     case 0x6b73:
          bsize = sizeof (struct sk_key);
          break;
     default:
          bsize = 0;
          result |= UNKNOWN_HEADER_BLOCK;
     }

     if ((goffset + 4 + bsize) > hdesc->size) {
          ALARM (TA_PROG_ERR, "structure for 0x%x "
                 "(size 0x%x) do not fit in file at offset 0x%"
                 PRIxOff, *id, bsize, goffset);
          return INVALID_HEADER_BLOCK;
     }
     if ((bsize + 4) > *seglen) {
          ALARM (TA_PROG_ERR, "structure for 0x%x "
                 "(size 0x%x) + size_header (4) do not fit in block "
                 "(size 0x%" PRId32 ") at offset 0x%" PRIxOff,
                 *id, bsize, *seglen, goffset);
          return INVALID_HEADER_BLOCK;
     }

     *infos = (void *) (hdesc->buffer + goffset + 4);

     return result;
}

void
classname_prn(struct hive *hdesc, struct nk_key *key)
{
     uint32_t boffset;
     
     if ((boffset = (uint32_t)LE32_TO_CPU(key->ofs_classnam)) != 0xFFFFFFFF ) {
          uint16_t cpos, clen = (uint16_t)LE16_TO_CPU(key->len_classnam);
          wchar_t mywchar;
	  int32_t bsize;
	  
          boffset += hdesc->h1offset;
	  if (get_block_size(hdesc, boffset, &bsize) != INVALID_HEADER_BLOCK) {
	       if (bsize < (clen + 4)) {
	            clen = bsize - 4;
	            ALARM(TA_PROG_ERR, "end of classname outside of block"
		         " at offset 0x%" PRIx32 ", keeping first %" PRIu16
			 " char(s)", boffset, clen/2);
	       }
	       boffset += 4;
	       for (cpos = 2; (boffset + 2) <= hdesc->size && cpos <= clen &&
	            (mywchar = DT_GETU16(hdesc->buffer + boffset)) != 0;
		    cpos += 2) {
	            dt_fputwc(FOutput, mywchar);
	            boffset += 2;
	       }
	  }
     }

     fputc ('|', FOutput);
}

int
root_path_prn (struct hive *hdesc)
{
     if (hdesc->root_path == NULL)
          return 0;

     fputs(hdesc->root_path, FOutput);
     return 1;
}

void
nk_name_prn (struct hive *hdesc, struct nk_key *key, int32_t block_size,
             int *valid, int *depth, off_t leaf)
{
     int i;
     uint16_t num_char, ktype;

     if (*valid == 0 && !hdesc->dump_any)
          return;

     if ((ktype = (uint16_t) LE16_TO_CPU (key->type)) == KEY_ROOT) {
          fprintf (FOutput, "0x%" PRIxOff "|%d|[", leaf, *valid);
     } else {
          if (walkback (hdesc,
                         (uint32_t) LE32_TO_CPU (key->ofs_parent) +
                         hdesc->h1offset, valid, depth, leaf) == -1) {
               *valid = 0;
               if (!hdesc->dump_any)
                    return;
               fprintf (FOutput, "0x%" PRIxOff "|%d|[", leaf, *valid);
          }
     }

     if (ktype == KEY_ROOT && hdesc->guess_root_path && root_path_prn(hdesc))
          return;

     num_char = (uint16_t) LE16_TO_CPU (key->len_name);

     if ((((void *) key->keyname) + num_char) >
         (((void *) key) + block_size - 4)) {
          num_char =
               ((void *) key) + block_size - 4 - ((void *) key->keyname);
          ALARM (TA_PROG_ERR,
                 "end of name outside of block for"
                 " key node at offset 0x%" PRIxOff ": keeping first %"
		 PRIu16 " char(s)",
                 leaf, num_char);
     }
     fputc ('\\', FOutput);
     for (i = 0; i < num_char && key->keyname[i]; i++)
          fputc ((int)key->keyname[i], FOutput);

}

int
walkback (struct hive *hdesc, off_t goffset, int *valid, int *depth,
          off_t leaf)
{
     int32_t block_size = 0;
     uint16_t block_id = 0;
     int hcheck = 0;
     struct nk_key *key;

     hcheck = check_block_header (hdesc, goffset, &block_size,
                                  &block_id, (void **) &key);
     if (hcheck == INVALID_HEADER_BLOCK)
          return -1;

     if (block_id != 0x6b6e) {
          ALARM (TA_PROG_ERR, "key node block expected instead of"
                 " type 0x%x at offset 0x%" PRIxOff, block_id, goffset);
          return -1;
     }

     if (hcheck & UNUSED_HEADER_BLOCK)
          *valid = 0;

     if ((*depth)++ > MAX_RECURSE_DEPTH) {
          ALARM (TA_PROG_ERR, "probable loop detected at offset 0x%"
                 PRIxOff, goffset);
          return -1;
     }

     nk_name_prn (hdesc, key, block_size, valid, depth, leaf);

     return 0;
}

void
sk_sd_prn (struct hive *hdesc, off_t goffset, off_t nk_offset)
{
     int32_t block_size = 0;
     uint16_t block_id = 0;
     uint32_t sd_size = 0;
     struct sk_key *key;

     if (check_block_header (hdesc, goffset, &block_size, &block_id,
                             (void **) &key) == INVALID_HEADER_BLOCK ||
         block_id != 0x6b73) {
          ALARM (TA_PROG_ERR, "can't find sk key header at offset"
                 " 0x%" PRIxOff ", id found 0x%" PRIx16, goffset, block_id);
          fputs ("||||", FOutput);
          return;
     }
     goffset += 0x18;
     sd_size =
          min ((off_t) (hdesc->size - goffset),
               (uint32_t) LE32_TO_CPU (key->len_sk));
     DumpBufferSecurity (FOutput, (uint64_t) nk_offset,
                         hdesc->buffer + goffset, (uint64_t) sd_size, 1);
     return;
}

int32_t
process_block (struct hive * hdesc, off_t goffset)
{
     int hcheck = 0, valid = 1, depth = 0;
     int32_t block_size;
     uint16_t block_id;
     struct nk_key *key;

     hcheck = check_block_header (hdesc, goffset, &block_size,
                                  &block_id, (void **) &key);
     if (hcheck == INVALID_HEADER_BLOCK)
          return 0;

     if (block_id != 0x6b6e) {
          if (FDbgOut) {
               fprintf (FDbgOut, "process_block: block of potential type 0x%x"
	                " at offset0x%" PRIxOff " is not a key node: "
			"skipping\n",
                        block_id, goffset);
          }
          return block_size;
     }

     if (hcheck & UNUSED_HEADER_BLOCK) {
          valid = 0;
          if (!hdesc->dump_any)
               return block_size;
     }

     /* fprintf(FOutput, "0x%"PRIxOff"|", goffset); */

     nk_name_prn (hdesc, key, block_size, &valid, &depth, goffset);

     if (valid == 0 && !hdesc->dump_any)
          return block_size;
	  
     fputs ("]|", FOutput);

     classname_prn(hdesc,key);

     time_prn (FOutput, (uint64_t) LE64_TO_CPU (key->timestamp));

     fprintf (FOutput, "|%" PRIu32 "|%" PRIu32 "|",
              (uint32_t) LE32_TO_CPU (key->no_subkeys),
              (uint32_t) LE32_TO_CPU (key->no_values));

/* add sk node processing to get security info... */
     sk_sd_prn (hdesc, (uint32_t) LE32_TO_CPU (key->ofs_sk) + hdesc->h1offset,
                goffset);

     fputc ('\n', FOutput);

     return block_size;
}

void
root_path_init(struct hive *hdesc)
{
     if (wcsncmp(hdesc->hive_name,
          L"\\SystemRoot\\System32\\Config\\SAM", HIVE_NAME_LENGTH) == 0) {
	  hdesc->root_path = "HKLM\\SAM";
     } else if (wcsncmp(hdesc->hive_name,
          L"SYSTEM", HIVE_NAME_LENGTH) == 0) {
          hdesc->root_path = "HKLM\\SYSTEM";
     } else if (wcsncmp(hdesc->hive_name,
          L"temRoot\\System32\\Config\\SOFTWARE", HIVE_NAME_LENGTH) == 0) {
	  hdesc->root_path = "HKLM\\SOFTWARE";
     } else if (wcsncmp(hdesc->hive_name,
          L"emRoot\\System32\\Config\\SOFTWARE", HIVE_NAME_LENGTH) == 0) {
	  hdesc->root_path = "HKLM\\SOFTWARE";
     } else if (wcsncmp(hdesc->hive_name,
          L"emRoot\\System32\\Config\\SECURITY", HIVE_NAME_LENGTH) == 0) {
	  hdesc->root_path = "HKLM\\SECURITY";
     } else if (wcsncmp(hdesc->hive_name,
          L"temRoot\\System32\\Config\\DEFAULT", HIVE_NAME_LENGTH) == 0) {
	  hdesc->root_path = "HKU\\.DEFAULT";
     } else {
          /* FIXME: any way to find user SID for NTUSER.DAT and
	   * UsrClass.dat ?
	   */
          ALARM (TA_PROG_ERR, "Failed to guess root path");
     }
}

void
parse_hive (struct hive *hdesc)
{
     off_t hoffset = HGLOBAL_OFFSET, goffset = 0;
     int32_t boffset = 0;
     uint32_t hsize = 0;
     struct regf_header *hdr;
     struct hbin_page *p;

     if (hdesc->size < sizeof (struct regf_header)) {
          ALARM (TA_PROG_ERR, "file %s is too small to be an hive",
                 hdesc->filename);
          return;
     }

     hdr = (struct regf_header *) hdesc->buffer;
     if (LE32_TO_CPU (hdr->id) != 0x66676572) {
          ALARM (TA_PROG_ERR, "file %s does not seem to be an hive",
                 hdesc->filename);
          return;
     }

     DumpSystemInfo (FOutput, "dirhive",
                     "Recorded Filename|Last Modification", 1);
     dt_get_unicode(hdr->name, hdesc->hive_name, HIVE_NAME_LENGTH);
     hdesc->hive_name[HIVE_NAME_LENGTH] = L'\0';
     dt_fputwcs(FOutput, hdesc->hive_name);
     fputc ('|', FOutput);
     time_prn (FOutput, (uint64_t) LE64_TO_CPU (hdr->timestamp));
     fputc ('\n', FOutput);
     fputs (FIELDSHEADERS, FOutput);
     if (hdesc->guess_root_path)
          root_path_init(hdesc);

     while ((hoffset + sizeof (struct hbin_page)) < hdesc->size) {
          p = (struct hbin_page *) (hdesc->buffer + hoffset);

          hsize = (uint32_t) LE32_TO_CPU (p->ofs_next);

          if ((uint32_t) LE32_TO_CPU (p->id) != 0x6e696268) {
               ALARM (TA_PROG_ERR, "hbin page signature not"
                      " found (0x%" PRIx32 ") at offset 0x%" PRIxOff
                      ", applying default offset",
                      LE32_TO_CPU (p->id), hoffset);
               hoffset += HGLOBAL_OFFSET;
               continue;
          }
          if (!hdesc->h1offset) {
               if (FDbgOut)
                    fprintf (FDbgOut, "parse_hive: first hbin block"
                             " found at 0x%" PRIxOff "\n", hoffset);
               hdesc->h1offset = hoffset;
          }

          if (LE32_TO_CPU (p->len_page) == 0) {
               ALARM (TA_PROG_WARN, "page len is null"
                      " for hbin page at offset 0x%" PRIxOff, hoffset);
          }

          if (hsize == 0) {
               ALARM (TA_PROG_ERR, "offset for next hbin page"
                      " at page at offset 0x%" PRIxOff " is null"
                      " using default size", hoffset);
               hsize = HGLOBAL_OFFSET;
          }

          if ((hoffset + hsize) > hdesc->size) {
               ALARM (TA_PROG_ERR, "end of hbin page at "
                      "offset 0x%" PRIxOff " after file"
                      " end, adjusting page size", hoffset);
               hsize = hdesc->size - hoffset;
          }

          goffset = hoffset + HBIN_HEADER_SIZE;

          while ((goffset - hoffset) < hsize) {
               if ((boffset = process_block (hdesc, goffset)) == 0) {
                    ALARM (TA_PROG_ERR, "at offset 0x%" PRIxOff
                           "while parsing hbin page at offset 0x%"
                           PRIxOff " skipping to next page",
                           goffset, hoffset);
                    break;
               }
               goffset += boffset;
          }

          hoffset += hsize;
     }
}

void
freeHive (struct hive *hdesc)
{
     if (hdesc->buffer) {
          DT_FREE (hdesc->buffer, "hive file buffer");
          hdesc->buffer = NULL;
     }

     if (hdesc->filedesc) {
          close (hdesc->filedesc);
          hdesc->filedesc = 0;
     }
}

int
openHiveFile (struct hive *hdesc)
{
     struct stat sbuf;
     ssize_t read_bytes;

     hdesc->buffer = NULL;
     hdesc->filedesc = 0;
     hdesc->filedesc = open (hdesc->filename, O_RDONLY);

     if (!hdesc->filedesc) {
          ALARM (TA_SYS_ERR, "unable to open filename %s", hdesc->filename);
          return 1;
     }

     if (fstat (hdesc->filedesc, &sbuf)) {
          ALARM (TA_SYS_ERR, "unable to stat the file %s", hdesc->filename);
          freeHive (hdesc);
          return 1;
     }

     hdesc->size = sbuf.st_size;

     if (hdesc->size > SIZE_MAX) {
          ALARM (TA_PROG_ERR, "file %s is too large to handle",
                 hdesc->filename);
          freeHive (hdesc);
          return 1;
     }

     hdesc->buffer = DT_CALLOC (char, (size_t) hdesc->size,
                                "hive file buffer (size %" PRIuOff ")",
                                hdesc->size);

#ifdef _WIN32
     _setmode (hdesc->filedesc, _O_BINARY);     /* to avoid CRLF conversion */
#endif

     read_bytes = read (hdesc->filedesc, hdesc->buffer, (size_t) hdesc->size);
     if (read_bytes < hdesc->size) {
          ALARM (TA_SYS_ERR,
                 "only %u out of %" PRIuOff " bytes read from file %s",
                 read_bytes, hdesc->size, hdesc->filename);
          freeHive (hdesc);
          return 1;
     }

     return 0;
}

int
openHiveStdin (struct hive *hdesc)
{
     size_t buff_size = 0, read_bytes = 0;
     void *ptr_realloc;

     hdesc->buffer = NULL;
     hdesc->filedesc = 0;
     hdesc->size = 0;

#ifdef _WIN32
     _setmode (_fileno (stdin), _O_BINARY);     /* to avoid CRLF conversion */
#endif
     do {
          buff_size += HALLOC_SIZE;
          ptr_realloc = DT_REALLOC (hdesc->buffer, char, buff_size,
                                    "hive buffer (size %u)", buff_size);
          hdesc->buffer = ptr_realloc;
          read_bytes =
               fread ((void *) (hdesc->buffer + hdesc->size), 1, HALLOC_SIZE,
                      stdin);
          hdesc->size += read_bytes;
     } while (!feof (stdin) && (read_bytes == HALLOC_SIZE) &&
              (hdesc->size <= (SIZE_MAX - HALLOC_SIZE)));
     hdesc->filename = stdin_filename;

     return 0;
}

void
usage (char *progname)
{
#ifdef _WIN32
     fprintf (stderr, "Usage: %s [-h][-a][-g][-8][-u user_file]"
              "[-d debug_file][-x error_file][-o file_output][filename]\n",
	      progname);
#else
     fprintf (stderr, "Usage: %s [-h][-a][-g][-8][-d [debug_file]]"
              "[-x error_file][-o file_output][filename]\n", progname);
#endif
}

void
help (char *progname)
{
     fprintf (stderr,
              "%s: list informations about keys inside an NT "
              "registry hive\n\n",
              progname);
     usage (progname);
     fprintf (stderr, "\nPackage version: %s\n", PACKAGE_STRING);
     fprintf (stderr, "\n\t--help, -h\tthis help\n");
     fprintf (stderr,
              "\t--any, -a\tlist any node found, instead of used ones\n");
     fprintf (stderr, "\t--guess-root-path, -g\ttry to figure out the complete "
          "root path (i.e. print \"HKLM\\SAM\" instead of \"SAM\")\n");
     fprintf (stderr,
	      "\t--utf8, -8\toutput non-ascii chars as UTF-8\n");
#ifdef _WIN32
     fprintf (stderr,
              "\t--user, -u\tsid to fullname correspondances\n");
#endif
     fprintf (stderr,
              "\t--debug, -d\tdebug mode");
#ifdef _WIN32
     fprintf (stderr, "\n");
# else
     fprintf (stderr, " (default output to stderr)\n");
#endif
     fprintf (stderr,
              "\t--error-output, -x\tsend errors to file instead of stderr\n");
     fprintf (stderr,
              "\t--output, -o\tsend output to file instead of stdout\n");
     fprintf (stderr,
              "\tif no registry file is specified on the command "
              "line, reads stdin\n");
     fprintf (stderr, "\nOutput format :\n");
     fprintf (stderr, FIELDSHEADERS);
     sd_help (stderr);
}

int
main (int argc, char **argv)
{
     struct hive hdesc;
     extern char *optarg;
     extern int optind, opterr, optopt;
     int c, read_success;
     char *pOutputFilename = NULL, *pDebugFilename = NULL, *pErrorFilename =
          NULL;
#ifdef _WIN32
     char *pUserFilename = NULL;
#endif
     DT_LOCALE_TYPE mylocale = DT_SET_LOCALE;
     static struct option long_opts[] = {
          {"help", 0, 0, 'h'},
          {"any", 0, 0, 'a'},
	  {"guess-root-path", 0, 0, 'g'},
	  {"utf8", 0, 0, '8'},
#ifdef _WIN32
          {"user", 1, 0, 'u'},
          {"debug", 1, 0, 'd'},
#else
          {"debug", 2, 0, 'd'},
#endif
          {"error-output", 1, 0, 'x'},
          {"output", 1, 0, 'o'},
          {0, 0, 0, 0}
     };
#ifdef _WIN32
#define OPTSTRING "hag8u:d:x:o:"
#else
#define OPTSTRING "hag8d::x:o:"
#endif

     FOutput = stdout;
     FDbgOut = NULL;
     FErrorOutput = stderr;
     hdesc.dump_any = 0;
     hdesc.guess_root_path =0;
     hdesc.hive_name[0] = L'\0';
     hdesc.root_path = NULL;
     hdesc.h1offset = 0;

     dt_setlocale("");
     while ((c = getopt_long (argc, argv, OPTSTRING, long_opts, NULL)) > 0) {
          switch (c) {
          case 'h':
               help (argv[0]);
               exit (EXIT_SUCCESS);
               break;
          case 'a':
               hdesc.dump_any = 1;
               break;
	  case 'g':
	        hdesc.guess_root_path = 1;
		break;
	  case '8':
	       mylocale = DT_SET_UTF8;
	       break;
#ifdef _WIN32
          case 'u':
	       MANDATORYFILENAME (optarg, pUserFilename, "user");
               break;
#endif
          case 'o':
               MANDATORYFILENAME (optarg, pOutputFilename, "output");
               break;
          case 'x':
               MANDATORYFILENAME (optarg, pErrorFilename, "error");
               break;
          case 'd':
               if (optarg)
                    pDebugFilename = optarg;
               else
                    FDbgOut = stderr;
               break;
          default:
               usage (argv[0]);
               exit (EXIT_FAILURE);
               break;
          }
     }

     set_wc_output_function(mylocale);

     if (optind == argc) {
          read_success = openHiveStdin (&hdesc);
     } else if (optind == (argc - 1)) {
          hdesc.filename = argv[optind];
          if (!hdesc.filename || !*hdesc.filename) {
               usage (argv[0]);
               exit (EXIT_FAILURE);
          }
          read_success = openHiveFile (&hdesc);
     } else {
          usage (argv[0]);
          exit (EXIT_FAILURE);
     }

     if (read_success)
          exit (EXIT_FAILURE);

     OUTPUTFILE (pOutputFilename, FOutput, "output");
     OUTPUTFILE (pErrorFilename, FErrorOutput, "error");
     OUTPUTFILE (pDebugFilename, FDbgOut, "debug");
#ifdef _WIN32
     FSecurDebug = FDbgOut;
     OUTPUTFILE (pUserFilename, FUserFile, "user");
#endif
     parse_hive (&hdesc);

     freeHive (&hdesc);

     exit (EXIT_SUCCESS);
}
