/******************************************************************************
 *
 * diskio.c - NTFS partition I/O. Part of the DirTools project.
 *
 * $Revision: 1.3 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include <inttypes.h>
#ifndef _WIN32
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <wchar.h>
#endif /* _WIN32 */
#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif /* USE_DMALLOC */

#include "diskio.h"
#include "dt_memory.h"
#include "dt_error.h"
#include "macros.h"
#include "output.h"
#include "md5.h"
#include "dt_locale.h"

#undef COMPRESSED_STREAMS_SUPPORT

FILE *FIoDebug = NULL;
int BIoDump = FALSE;

uint32_t DwMaxRead;             /* Maximum number of bytes to be read
                                 * from raw disk */
uint32_t DwProcessStreamBuffer; /* Buffer size to process streams */
/* Various buffer needed */
unsigned char *PProcessStreamBuffer = NULL;
#ifdef COMPRESSED_STREAMS_SUPPORT
unsigned char *PChunkBufferIn = NULL;   /* Buffer to read chunks
                                         * (compression unit = 16 clusters) */
unsigned char *PChunkBufferOut = NULL;  /* Buffer to output chunks */
#endif /* COMPRESSED_STREAMS_SUPPORT */
unsigned char *PClusterBuffer = NULL;

/**********************************************************************
 * ReadClusters : fill a buffer with clusters read from raw disk
 *
 * - [in] pVol : pointer to volume description structure
 * - [in] ulOffset : cluster offset from the beginning of the volume
 * - [in/out] pNumber : pointer to the number of clusters requested/read
 * - [out] pBuffer : pointer to the buffer for read clusters
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
int
ReadClusters (PNTFS_VOL_INFO pVol, uint64_t ulOffset,
              uint64_t * pNumber, unsigned char *pBuffer)
{
#ifdef _WIN32
     ULARGE_INTEGER ulStartNum;
#else
     ssize_t ssResult;
#endif /* _WIN32 */
     uint64_t ulSize, ulRead;
     uint32_t dwResult;

     ulSize = *pNumber * pVol->dwClusterSize;
     *pNumber = 0;
     if (ulOffset >= pVol->ulClusterNumber) {
          ALARM (TA_PROG_ERR, "offset beyond end of volume");
          return FALSE;
     }
#ifdef _WIN32
     ulStartNum.QuadPart = ulOffset * pVol->dwClusterSize;
     dwResult = SetFilePointer (pVol->hVolume,
                                ulStartNum.LowPart,
                                &ulStartNum.HighPart, FILE_BEGIN);
     if ((dwResult == INVALID_SET_FILE_POINTER) &&
         (GetLastError () != OK_ERROR)) {
          ALARM (TA_SYS_ERR, "SetFilePointer - Offset : 0x%" PRIx64,
                 ulStartNum.QuadPart);
          return FALSE;
     }
#else /* _WIN32 */
     if (lseek (pVol->hVolume, ulOffset * pVol->dwClusterSize + pVol->ulSkip,
                SEEK_SET) == -1) {
          ALARM (TA_SYS_ERR, "lseek - Offset : 0x%" PRIx64,
                 ulOffset * pVol->dwClusterSize);
          return FALSE;
     }
#endif /* _WIN32 */
     dwResult = 1;              /* Non null value */
     ulRead = 0;
     while (ulSize && dwResult) {
          dwResult = min (ulSize, DwMaxRead);
#ifdef _WIN32
          if (!ReadFile (pVol->hVolume,
                         pBuffer + ulRead, dwResult, &dwResult, NULL)) {
               ALARM (TA_SYS_ERR,
                      "ReadFile for %" PRIu32
                      " bytes failed - Already read bytes : %" PRIu64
                      ", remaining : %" PRIu64, dwResult, ulRead, ulSize);
               *pNumber = ulRead / pVol->dwClusterSize;
               return FALSE;
          }
          ulSize -= dwResult;
          ulRead += dwResult;

#else /* _WIN32 */
          if ((ssResult = read (pVol->hVolume,
                                pBuffer + ulRead, dwResult)) == -1) {
               ALARM (TA_SYS_ERR,
                      "read for %" PRIu32
                      " bytes failed - Already read bytes : %" PRIu64
                      ", remaining : %" PRIu64, dwResult, ulRead, ulSize);
               *pNumber = ulRead / pVol->dwClusterSize;
               return FALSE;
          }
          ulSize -= ssResult;
          ulRead += ssResult;
          dwResult = ssResult;

#endif /* _WIN32 */
     }

     *pNumber = ulRead / pVol->dwClusterSize;

     return TRUE;
}

/**********************************************************************
 * ExploreRunlist
 *
 * - [in] pAttr : pointer to attribute description structure
 * - [out] pIndex : pointer to the index position in the runlist
 * - [in/out] pOffset : pointer to the requested offset, returns
 *                      as an offset inside the indexed position found
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
int
ExploreRunlist (PNTFS_ATTRIBUTE pAttr, uint64_t * pIndex, uint64_t * pOffset)
{
     *pIndex = 0;
     while (*pIndex < pAttr->ulRunlistSize) {
          if (*pOffset < pAttr->pRunlist[*pIndex].ulLength)
               return TRUE;

          *pOffset -= pAttr->pRunlist[*pIndex].ulLength;
          (*pIndex)++;
     }
     return FALSE;
}

/**********************************************************************
 * ReadStreamVcns : fill a buffer with the clusters of an attribute stream
 *
 * - [in] pVol : pointer to volume description structure
 * - [in] pAttr : pointer to attribute description structure
 * - [in] ulStartNum : offset to the first vcn required
 * - [in/out] pNumber : pointer to number of clusters required/read
 * - [out] pBuffer : pointer to the buffer for read clusters
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
int
ReadStreamVcns (PNTFS_VOL_INFO pVol, PNTFS_ATTRIBUTE pAttr,
                uint64_t ulStartNum, uint64_t * pNumber,
                unsigned char *pBuffer)
{
     uint64_t ulLeft, ulRead, ulMemo, ulIndex;

     if (pAttr->bResident) {
          ALARM (TA_PROG_ERR, "do not apply to resident streams");
          *pNumber = 0;
          return FALSE;
     }
     /* Runlist cases... */
     if (pAttr->wCompressed) {
          /* Runlist compressed */
          ALARM (TA_PROG_ERR, "compressed attributes not implemented");
          *pNumber = 0;
          return FALSE;
     }
     /* Runlist not compressed */
     ulLeft = *pNumber;
     if (ExploreRunlist (pAttr, &ulIndex, &ulStartNum)) {
          while (ulLeft && (ulIndex < pAttr->ulRunlistSize)) {
               ulRead =
                    min (ulLeft,
                         pAttr->pRunlist[ulIndex].ulLength - ulStartNum);
               if (pAttr->pRunlist[ulIndex].ulLcn == NTFS_VIRTUAL_LCN) {
#ifdef _WIN32
                    ZeroMemory (pBuffer, ulRead);
#else
                    /* FIXME : dangerous cast, loop instead ! */
                    memset (pBuffer, 0, (size_t) ulRead);
#endif
                    ulLeft -= ulRead;
                    ulIndex++;
                    ulStartNum = 0;
               } else {
                    ulMemo = ulRead;
                    ReadClusters (pVol,
                                  ulStartNum + pAttr->pRunlist[ulIndex].ulLcn,
                                  &ulRead, pBuffer);
                    ulLeft -= ulRead;
                    ulIndex++;
                    ulStartNum = 0;
                    if (ulRead < ulMemo) {
                         ALARM (TA_PROG_ERR,
                                "(attribute 0x%" PRIx32 ") only %" PRIu64
                                " out of %" PRIu64 " clusters read at Lcn 0x%"
                                PRIx64, pAttr->dwType, ulRead, ulMemo,
                                pAttr->pRunlist[ulIndex].ulLcn);
                         *pNumber -= ulLeft;
                         return FALSE;
                    }
               }
               pBuffer += ulRead * pVol->dwClusterSize;
          }                     /* while (ulLeft && (ulIndex < pAttr->... */
     }                          /* if (ExploreRunlist(pAttr, &ulIndex,... */
     if (ulLeft) {
          ALARM (TA_PROG_ERR,
                 "(attribute 0x%" PRIx32 ") %" PRIu64
                 " clusters not found in runlist", pAttr->dwType, ulLeft);
          return FALSE;
     }

     return TRUE;
}

#ifdef COMPRESSED_STREAMS_SUPPORT
/**********************************************************************
 * DecompressChunk : fill a buffer with the chunks of an attribute stream
 *
 * - [in] pVol : pointer to volume description structure
 * - [in] pBufferIn : pointer to compressed buffer start
 * - [in] pMax : pointer to compressed buffer end + 1
 * - [out] pBufferOut : pointer to the buffer for read chunks
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
int
DecompressChunk (PNTFS_VOL_INFO pVol, unsigned char *pBufferIn,
                 unsigned char *pMax, unsigned char *pBufferOut)
{
#define COMPRESSION_UNIT 4096
     uint16_t wSize;

     while (TRUE) {
          if ((pBufferIn + 1) < pMax) {
               wSize = DT_GETU16 (pBufferIn);
               pBufferIn += 2;
          } else
               break;

          if (((wSize & 0xFFF) < 0xFFF) && !(wSize & 0x8000)) {
               /* The compression bit is not set and the size is
                * less than an uncompressed unit... */
               ALARM (TA_PROG_ERR,
                      "wrong compression unit size tag : %" PRIu16, wSize);
               return FALSE;
          }

          wSize = (wSize & 0xFFF) + 1;
          if ((wSize + pBufferIn) > pMax) {
               ALARM (TA_PROG_ERR,
                      "missing bytes in compression unit %" PRIu16
                      " required, %" PRIu16 " found", wSize,
                      (uint16_t) (pMax - pBufferIn));
               return FALSE;
          }

          if (wSize == COMPRESSION_UNIT) {
#ifdef _WIN32
/* Tester la valeur de retour */
               CopyMemory (pBufferOut, pBufferIn, COMPRESSION_UNIT);
#else /* _WIN32 */
               memcpy (pBufferOut, pBufferIn, COMPRESSION_UNIT);
#endif /* _WIN32 */
               pBufferOut += COMPRESSION_UNIT;
               pBufferIn += COMPRESSION_UNIT;
          } else {
          }
     }

     return TRUE;
}

/**********************************************************************
 * ReadStreamChunks : fill a buffer with the chunks of an attribute stream
 *
 * - [in] pVol : pointer to volume description structure
 * - [in] pAttr : pointer to attribute description structure
 * - [in] ulFirstChunk : offset to the first chunk required
 * - [in/out] pChunkNumber : pointer to number of chunks required/read
 * - [out] pBuffer : pointer to the buffer for read chunks
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
int
ReadStreamChunks (PNTFS_VOL_INFO pVol, PNTFS_ATTRIBUTE pAttr,
                  uint64_t ulFirstChunk, uint64_t * pChunkNumber,
                  unsigned char *pBuffer)
{
     uint64_t ulChunkLeft, ulStart, ulIndex = 0, ulRead, ulMemo;
     uint16_t wClusterLeft;
     unsigned char *pTempo = PChunkBufferIn;

     if (pAttr->bResident) {
          ALARM (TA_PROG_ERR, "do not apply to resident streams");
          *pChunkNumber = 0;
          return FALSE;
     }

     if ((!pAttr->wCompressed) || (pAttr->wCompEngine != 4)) {
          ALARM (TA_PROG_ERR,
                 "not compressed or unknown compression engine : 0x%" PRIx16,
                 pAttr->wCompEngine);
          *pChunkNumber = 0;
          return FALSE;
     }

     ulChunkLeft = *pChunkNumber;
     ulStart = ulFirstChunk * CHUNK_CLUSTER_SIZE;
     if (ExploreRunlist (pAttr, &ulIndex, &ulStart)) {
          wClusterLeft = CHUNK_CLUSTER_SIZE;
          while (ulChunkLeft && (ulIndex < pAttr->ulRunlistSize)) {
               ulRead =
                    min (wClusterLeft,
                         pAttr->pRunlist[ulIndex].ulLength - ulStart);
               if (pAttr->pRunlist[ulIndex].ulLcn == NTFS_VIRTUAL_LCN) {
                    if (ulRead != wClusterLeft) {
                         *pChunkNumber -= ulChunkLeft;
                         ALARM (TA_PROG_ERR,
                                "%" PRIu16 " virtual cluster(s)"
                                " expected instead of %" PRIu64 " at Vcn 0x%"
                                PRIx64, wClusterLeft, ulRead,
                                (ulFirstChunk +
                                 *pChunkNumber) * CHUNK_CLUSTER_SIZE);
                         return FALSE;
                    }

                    /* In case of sparse unit (a whole chunk made of
                     * virtual clusters)
                     * Fill the buffer with a chunk full of zeros */
                    if (wClusterLeft == CHUNK_CLUSTER_SIZE) {
#ifdef _WIN32
                         if (!ZeroMemory (pBuffer, pVol->dwChunkSize)) {
                              *pChunkNumber -= ulChunkLeft;
                              ALARM (TA_SYS_ERR, "ZeroMemory");
                              return FALSE;
                         }
#else /* _WIN32 */
                         memset (pBuffer, 0, pVol->dwChunkSize);
#endif /* _WIN32 */
                         pBuffer += ulRead * pVol->dwClusterSize;
                         ulChunkLeft--;
                    } else      /* if (wClusterLeft == CHUNK_CLUSTER_SIZE) */
                         wClusterLeft -= ulRead;        /* should give
                                                         * wClusterLeft = 0 */
               } else {         /* if (pAttr->pRunlist[ulIndex].ulLcn ==...*/

                    ulMemo = ulRead;
                    ReadClusters (pVol,
                                  ulStart + pAttr->pRunlist[ulIndex].ulLcn,
                                  &ulRead, pTempo);
                    wClusterLeft -= ulRead;
                    if (ulRead < ulMemo) {
                         *pChunkNumber -= ulChunkLeft;
                         ALARM (TA_PROG_ERR,
                                "ReadStreamChunks : only %" PRIu64 " out of %"
                                PRIu64 " clusters read at Lcn 0x%" PRIx64,
                                ulRead, ulMemo,
                                pAttr->pRunlist[ulIndex].ulLcn);
                         return FALSE;
                    }
                    pTempo += ulRead * pVol->dwClusterSize;
               }                /* if (pAttr->pRunlist[ulIndex].ulLcn ==...*/

               ulStart += ulRead;

               if (!wClusterLeft) {
                    ulChunkLeft--;
                    pTempo = PChunkBufferIn;
                    wClusterLeft = CHUNK_CLUSTER_SIZE;
                    /* Dcompression */
               }
               if (ulStart == pAttr->pRunlist[ulIndex].ulLength) {
                    ulStart = 0;
                    ulIndex++;
               }
          }                     /* while(ulChunkLeft && (ulIndex..*/

     }
     /* if (ExploreRunlist(pAttr, &ulIndex, &ulStart)) */
     if (ulChunkLeft) {
          ALARM (TA_PROG_ERR, "%" PRIu64 " chunks not found in runlist",
                 ulChunkLeft);
          return FALSE;
     }

     return TRUE;
}
#endif /* COMPRESSED_STREAMS_SUPPORT */

/**********************************************************************
 * ReadStreamBytes : fill a buffer with the bytes of an attribute stream
 *
 * - [in] pVol : pointer to volume description structure
 * - [in] pAttr : pointer to attribute description structure
 * - [in] ulOffset : offset to the first byte required
 * - [in/out] pSize : pointer to number of bytes required/read
 * - [out] pBuffer : pointer to the buffer for read bytes
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
int
ReadStreamBytes (PNTFS_VOL_INFO pVol, PNTFS_ATTRIBUTE pAttr,
                 uint64_t ulOffset, uint64_t * pSize, unsigned char *pBuffer)
{
     uint32_t dwLength;
     uint64_t ulStartVcn, ulLeft, ulClusterNumber;

     if (ulOffset > pAttr->ulStreamSize) {
          ALARM (TA_PROG_ERR,
                 "Offset (0x%" PRIx64 ") > stream size (0x%" PRIx64 ")",
                 ulOffset, pAttr->ulStreamSize);
          *pSize = 0;
          return FALSE;
     }
     if (pAttr->bResident) {
          dwLength = min (*pSize, pAttr->ulStreamSize - ulOffset);
#ifdef _WIN32
          if (!CopyMemory (pBuffer, pAttr->pStream + ulOffset, dwLength)) {
               ALARM (TA_SYS_ERR, "CopyMemory resident stream - "
                      "Requested length : %" PRIu32, dwLength);
               *pSize = 0;
               return FALSE;
          }
#else /* _WIN32 */
          memcpy (pBuffer, pAttr->pStream + ulOffset, dwLength);
#endif /* _WIN32 */
          *pSize = dwLength;
     } else {                   /* if (pAttrib->bResident) */

          /* Runlist cases... */
          if (pAttr->wCompressed) {
               /* Runlist compressed */
               ALARM (TA_PROG_ERR, "compressed attributes not implemented");
               *pSize = 0;
               return FALSE;
          } else {              /* if (pAttrib->wCompressed) */

               /* Runlist not compressed */
               ulStartVcn = ulOffset / pVol->dwClusterSize;
               ulOffset = ulOffset % pVol->dwClusterSize;
               ulLeft = *pSize;
               while (ulLeft) {
                    if (ulOffset
                        || ((ulOffset + ulLeft) < pVol->dwClusterSize)) {
                         /* if we only want part of some clusters,
                          * use PClusterBuffer
                          * to avoid pBuffer overflow... */
                         dwLength =
                              min (pVol->dwClusterSize - ulOffset, ulLeft);
                         ulClusterNumber = 1;
                         if (!ReadStreamVcns (pVol,
                                              pAttr,
                                              ulStartVcn,
                                              &ulClusterNumber,
                                              PClusterBuffer)) {
                              *pSize -= ulLeft;
                              return FALSE;
                         }
#ifdef _WIN32
                         if (!CopyMemory
                             (pBuffer, PClusterBuffer + ulOffset, dwLength)) {
                              ALARM (TA_SYS_ERR,
                                     "CopyMemory from PClusterBuffer"
                                     " - Requested length : %" PRIu32,
                                     dwLength);
                              *pSize -= ulLeft;
                              return FALSE;
                         }
#else /* _WIN32 */
                         memcpy (pBuffer, PClusterBuffer + ulOffset,
                                 dwLength);
#endif /* _WIN32 */
                         pBuffer += dwLength;
                         ulLeft -= dwLength;
                         ulStartVcn += ulClusterNumber;
                         ulOffset = 0;
                    } else {    /* if (ulOffset || ((ulOffset + ulLeft) ... */

                         ulClusterNumber = ulLeft / pVol->dwClusterSize;
                         if (!ReadStreamVcns (pVol,
                                              pAttr,
                                              ulStartVcn,
                                              &ulClusterNumber, pBuffer)) {
                              *pSize -= ulLeft;
                              return FALSE;
                         }
                         pBuffer += ulClusterNumber * pVol->dwClusterSize;
                         ulLeft -= ulClusterNumber * pVol->dwClusterSize;
                         ulStartVcn += ulClusterNumber;
                    }           /* if (ulOffset || ((ulOffset + ulLeft) <... */
               }                /* while (ulLeft) */
          }                     /* if (pAttrib->wCompressed) */
     }                          /* if (pAttrib->bResident) */
     return TRUE;
}

/**********************************************************************
 * ReleaseStreamBuffer : free (if needed) the buffer of the stream
 *                       of a given attribute
 *
 * - [in] pAttr : pointer to attribute description structure
 * - [in] pBuffer : pointer to the buffer to free
 *
 * returns OK_ERROR in case of success, FATAL_ERROR otherwise
 *
 *********************************************************************/
int
ReleaseStreamBuffer (PNTFS_ATTRIBUTE pAttr, unsigned char *pBuffer)
{
     if ((!pAttr->bResident) && pBuffer) {
          DT_FREE (pBuffer, "pBuffer");
     }
     return OK_ERROR;
}

/**********************************************************************
 * GetStreamBuffer : allocate and fill (if needed) a buffer to an
 *                   attribute stream
 *
 * - [in] pEntry : pointer to entry (file description) structure
 * - [in] pAttr : pointer to attribute description structure
 * - [out] pSize : pointer to the size of the buffer
 *
 * returns a pointer to the (allocated) buffer, NULL otherwise
 *
 *********************************************************************/
unsigned char *
GetStreamBuffer (PNTFS_MFT_ENTRY pEntry, PNTFS_ATTRIBUTE pAttr,
                 uint64_t * pSize)
{
     unsigned char *pResult = NULL;

     /* Get the stream size collected in the header part */
     *pSize = pAttr->ulStreamSize;

     if (pAttr->bResident) {
          /* If it is a resident stream, the stream buffer is already
           * within the structure */
          pResult = pAttr->pStream;
     } else {                   /* if (pAttr->bResident) */

          /* Allocate the memory needed */
          pResult = DT_CALLOC (unsigned char, *pSize, "base record #%" PRIu64
                               ", attribute type 0x%" PRIx32,
                               pEntry->ulNumber, pAttr->dwType);
          /* Fill the buffer with the non-resident stream */
          ReadStreamBytes (pEntry->pVolume, pAttr, 0, pSize, pResult);

     }                          /* if (pAttr->bResident) */

     return pResult;
}

/**********************************************************************
 * ProcessStream : apply a function to the stream of an attribute
 *
 * - [in] pEntry : pointer to entry (file description) structure
 * - [in] pAttr : pointer to attribute description structure
 * - [in] pFunc : pointer to the function to apply to the stream
 * - [in] pFuncArg : argument specific to the function used
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
int
ProcessStream (PNTFS_MFT_ENTRY pEntry, PNTFS_ATTRIBUTE pAttr,
               void (*pFunc) (), void *pFuncArg)
{
     uint64_t ulOffset = 0;
     uint64_t ulRead;

     if (pAttr->bResident) {
          (*pFunc) (pFuncArg, pAttr->pStream, (uint32_t) pAttr->ulStreamSize);  /* Max resident stream size 4Ko */
          return TRUE;
     }

     if (!PProcessStreamBuffer)
          PProcessStreamBuffer =
               DT_CALLOC (unsigned char, DwProcessStreamBuffer,
                          "PProcessStreamBuffer");

     while (ulOffset < pAttr->ulStreamSize) {
          ulRead = min ((uint64_t) DwProcessStreamBuffer,
                        pAttr->ulStreamSize - ulOffset);
          ReadStreamBytes (pEntry->pVolume, pAttr, ulOffset, &ulRead,
                           PProcessStreamBuffer);
          ulOffset += ulRead;
          if (!ulRead)
               break;

          (*pFunc) (pFuncArg, PProcessStreamBuffer, (uint32_t) ulRead);
          /* ulRead <= 1 Mo */
     }
     if (ulOffset < pAttr->ulStreamSize)
          return FALSE;
     else
          return TRUE;
}

/**********************************************************************
 * FindAttribute : find an attribute of a given type and eventually
 *                 name, in an entry structure
 *
 * - [in] pEntry : pointer to entry (file description) structure
 * - [in] dwType : type to look for
 * - [in] pName : pointer to the name to look for.
 *                Ignored if NULL
 *
 * returns a pointer to the attribute, NULL if not found
 *
 *********************************************************************/
PNTFS_ATTRIBUTE
FindAttribute (PNTFS_MFT_ENTRY pEntry, uint32_t dwType, wchar_t *pName)
/* If pName == NULL, only look for type */
{
     PNTFS_ATTRIBUTE pResult = NULL;
     uint16_t wCount;

     if (!pEntry->wAttrCount)
          return pResult;

     for (wCount = 0; wCount < pEntry->wAttrCount; wCount++) {
          if (pEntry->pAttr[wCount].dwType == dwType) {
               if (pName) {
                    if (!wcsncmp (pEntry->pAttr[wCount].wcName,
                                  pName, ATTRIBUTE_MAX_NAME_LENGTH))
                    {
                         pResult = &pEntry->pAttr[wCount];
                         break;
                    }
               } else {
                    pResult = &pEntry->pAttr[wCount];
                    break;
               }
          }
     }

     return pResult;
}

/**********************************************************************
 * ProcessRunlist : put in an attribute structure the runlist read
 *                  from a buffer using NTFS byte-order
 *
 * - [in/out] pAttr : pointer to the attribute stucture to complete
 * - [in] pData : pointer to the buffer to read from
 * - [in] ulStartVcn : offset to the first vcn in the runlist
 * - [in] ulEndVcn : last vcn in the runlist
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
#define RUNLIST_NUMBER_ALLOCATED 10
int
ProcessRunlist (PNTFS_ATTRIBUTE pAttr,
                unsigned char *pData, uint64_t ulStartVcn, uint64_t ulEndVcn)
{
     unsigned char byTag;
     uint32_t dwOffset = 0;
     PNTFS_RUNLIST pCopy = NULL;
     uint64_t ulLcn = 0;

     if (ulStartVcn != pAttr->ulVcnsNumber) {
          ALARM (TA_PROG_ERR, "missing Vcns - Expected : 0x%" PRIx64
                 " Current 0x%" PRIx64 " - Attribute type 0x%" PRIx32,
                 pAttr->ulVcnsNumber, ulStartVcn, pAttr->dwType);
          return FALSE;
     }

     if (!pAttr->pRunlist) {
          pAttr->pRunlist = DT_CALLOC (NTFS_RUNLIST, 1, "pRunlist");
     }

     /* Add 1 to get the number of Vcn instead of the last one */
     ulEndVcn++;
     /* Init the Lcn value with the first cluster of the last run
      * of the runlist, or 0 by default
      * THIS was a MISTAKE ! Each Runlist start at cluster 0
      * ulLcn = pAttr->ulLastRunLcn; */

     while (ulStartVcn < ulEndVcn) {
          if ((pAttr->ulRunlistSize % RUNLIST_NUMBER_ALLOCATED) == 0) {
               pCopy = DT_REALLOC (pAttr->pRunlist, NTFS_RUNLIST,
                                   pAttr->ulRunlistSize +
                                   RUNLIST_NUMBER_ALLOCATED,
                                   "pCopy - Attribute type 0x%" PRIx32,
                                   pAttr->dwType);
               pAttr->pRunlist = pCopy;
          }
          /* if ((pAttr->ulRunlistSize % 10) == 0) */
          byTag = DT_GETU8 (pData + dwOffset);
          dwOffset++;
          switch (byTag & 0xF) {
          case 1:
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLength =
                    DT_GETU8 (pData + dwOffset);
               dwOffset++;
               break;
          case 2:
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLength =
                    DT_GETU16 (pData + dwOffset);
               dwOffset += 2;
               break;
          case 3:
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLength =
                    DT_GETU24 (pData + dwOffset);
               dwOffset += 3;
               break;
          case 4:
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLength =
                    DT_GETU32 (pData + dwOffset);
               dwOffset += 4;
               break;
          case 5:
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLength =
                    DT_GETU40 (pData + dwOffset);
               dwOffset += 5;
               break;
          case 6:
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLength =
                    DT_GETU48 (pData + dwOffset);
               dwOffset += 6;
               break;
          case 7:
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLength =
                    DT_GETU56 (pData + dwOffset);
               dwOffset += 7;
               break;
          case 8:
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLength =
                    DT_GETU64 (pData + dwOffset);
               dwOffset += 8;
               break;
          default:
               /* Erreur ?? */
               fprintf (stderr, "FIXME : strange value %u for run offset\n",
                        byTag & 0xF);
               break;
          }                     /* switch (byTag & 0xF) */
          ulStartVcn += pAttr->pRunlist[pAttr->ulRunlistSize].ulLength;

          switch (byTag >> 4) {
          case 0:
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn = NTFS_VIRTUAL_LCN;
               break;
          case 1:
               ulLcn += DT_GETS8 (pData + dwOffset);
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn = ulLcn;
               dwOffset++;
               break;
          case 2:
               ulLcn += DT_GETS16 (pData + dwOffset);
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn = ulLcn;
               dwOffset += 2;
               break;
          case 3:
               ulLcn += DT_GETS24 (pData + dwOffset);
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn = ulLcn;
               dwOffset += 3;
               break;
          case 4:
               ulLcn += DT_GETS32 (pData + dwOffset);
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn = ulLcn;
               dwOffset += 4;
               break;
          case 5:
               ulLcn += DT_GETS40 (pData + dwOffset);
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn = ulLcn;
               dwOffset += 5;
               break;
          case 6:
               ulLcn += DT_GETS48 (pData + dwOffset);
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn = ulLcn;
               dwOffset += 6;
               break;
          case 7:
               ulLcn += DT_GETS56 (pData + dwOffset);
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn = ulLcn;
               dwOffset += 7;
               break;
          case 8:
               ulLcn += DT_GETS64 (pData + dwOffset);
               pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn = ulLcn;
               dwOffset += 8;
               break;
          default:
               /* Erreur ?? */
               fprintf (stderr, "FIXME : strange value %u for run length\n",
                        byTag >> 4);
               break;
          }                     /* switch (byTag >> 4) */
          /* If it is a real run (no virtual clusters)
           * memorize the first Lcn of the run */
          if (pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn != NTFS_VIRTUAL_LCN)
               pAttr->ulLastRunLcn =
                    pAttr->pRunlist[pAttr->ulRunlistSize].ulLcn;

          /* one more element in the runlist */
          pAttr->ulRunlistSize++;
     }                          /* while (ulStartVcn <= ulEndVcn) */

     pAttr->ulVcnsNumber = ulEndVcn;

     return TRUE;
}

/**********************************************************************
 * InitAttribute : add to an entry structure the attribute present
 *                 in a buffer using NTFS byte-order
 *
 * - [in/out] pEntry : pointer to entry (file description) structure
 * - [in] pData : pointer to buffer to read from
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
/* 7 attributes should be a good value : 1 Standard, 2 File Name,
 * 1 Security AND (1 Data OR (1 Index root, 1 Index allocation,
 * 1 Bitmap)) */
#define ATTRIBUTE_NUMBER_ALLOCATED 7

int
InitAttribute (PNTFS_MFT_ENTRY pEntry, unsigned char *pData)
{
     uint32_t dwType;
     uint16_t wNameLength;
     wchar_t wcName[ATTRIBUTE_MAX_NAME_LENGTH];
     PNTFS_ATTRIBUTE pAttr = NULL, pAttrList = NULL;

     dwType = DT_GETU32 (pData);
     wNameLength = DT_GETU8 (pData + 0x9);
     if (wNameLength)
          dt_get_unicode (pData + DT_GETU16 (pData + 0xA), wcName, wNameLength);

     wcName[wNameLength] = L'\0';

     /* I suppose only non-resident attributes
      * can be splitted among various records... */
     if (DT_GETU8 (pData + 0x8))
          pAttr = FindAttribute (pEntry, dwType, wcName);
     else
          pAttr = NULL;

     if (!pAttr)
          /* The attribute does not exist in the entry */
     {
          if (pEntry->pAttr) {
               /* There's already some attributes => resize the buffer */
               if ((pEntry->wAttrCount % ATTRIBUTE_NUMBER_ALLOCATED) == 0)
                    pAttrList = DT_REALLOC (pEntry->pAttr, NTFS_ATTRIBUTE,
                                            pEntry->wAttrCount +
                                            ATTRIBUTE_NUMBER_ALLOCATED,
                                            "pAttrList - Base record #%"
                                            PRIu64, pEntry->ulNumber);
               else
                    pAttrList = pEntry->pAttr;
          } else {              /* if (pEntry->pAttr) */

               /* No attributes in the entry => new buffer */
               pAttrList =
                    DT_CALLOC (NTFS_ATTRIBUTE, ATTRIBUTE_NUMBER_ALLOCATED,
                               "pAttrList - Base record #%" PRIu64,
                               pEntry->ulNumber);
          }                     /* if (pEntry->pAttr) */

          pEntry->pAttr = pAttrList;

          pAttr = &pAttrList[pEntry->wAttrCount];
          pAttr->dwType = dwType;
          pAttr->wNameLength = wNameLength;
          wcsncpy (pAttr->wcName, wcName, ATTRIBUTE_MAX_NAME_LENGTH);
          pAttr->bResident = (DT_GETU8 (pData + 0x8) == 0 ? TRUE : FALSE);
          pAttr->wCompressed = DT_GETU16 (pData + 0xC);
          /* Safe defaults */
          pAttr->ulStreamSize = 0;
          pAttr->pStream = NULL;
          pAttr->ulRunlistSize = 0;
          pAttr->ulVcnsNumber = 0;
          pAttr->ulLastRunLcn = 0;
          pAttr->pRunlist = NULL;

          if (FIoDebug) {
	       /* FIXME: dt_fputwcs */
               fprintf (FIoDebug,
                         "Adding attribute 0x%" PRIx32
                         ", name \"", pAttr->dwType);
	       dt_fputwcs(FIoDebug, pAttr->wcName);
	       fprintf(FIoDebug, "\", to base record #%" PRIu64 "\n",
                         pEntry->ulNumber);
	  }
	  
          if (pAttr->bResident) {
               pAttr->ulStreamSize = DT_GETU32 (pData + 0x10);
               pAttr->pStream = DT_CALLOC (unsigned char, pAttr->ulStreamSize,
                                           "pStream - Entry #%" PRIu64
                                           " attribute type 0x%" PRIx32,
                                           pEntry->ulNumber, pAttr->dwType);
#ifdef _WIN32
/* FIXME check value */
               CopyMemory (pAttr->pStream,
                           pData + DT_GETU16 (pData + 0x14),
                           pAttr->ulStreamSize);
#else /* _WIN32 */
               memcpy (pAttr->pStream,
                       pData + DT_GETU16 (pData + 0x14), pAttr->ulStreamSize);
#endif /* _WIN32 */
          } else {              /* if (pAttr->bResident) */

               pAttr->ulStreamSize = DT_GETU64 (pData + 0x30);
               pAttr->ulAllocated = DT_GETU64 (pData + 0x28);
               pAttr->ulCompressed = DT_GETU64 (pData + 0x38);
               pAttr->wCompEngine = DT_GETU16 (pData + 0x22);
               pAttr->ulRunlistSize = 0;
               pAttr->ulVcnsNumber = 0;
          }                     /* if (pAttr->bResident) */

          pEntry->wAttrCount++;
     }
     /* if (! pAttr) */
     if (DT_GETU8 (pData + 0x8)) {
          /* Non resident attribute : add the runlist */
/* FIXME Check return value */
          ProcessRunlist (pAttr,
                          pData + DT_GETU16 (pData + 0x20),
                          DT_GETU64 (pData + 0x10), DT_GETU64 (pData + 0x18));
     }
     return TRUE;

}

/**********************************************************************
 * ReadAllAttributes : add to an entry structure all the attributes
 *                    found in a buffer using NTFS byte-order
 *
 * - [in/out] pEntry : pointer to entry (file description) structure
 * - [in] pData : pointer to buffer to read from
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
int
ReadAllAttributes (PNTFS_MFT_ENTRY pEntry, unsigned char *pData)
{
     uint32_t dwSize, dwType, dwOffset;

     dwOffset = DT_GETU16 (pData + 0x14);
     dwSize = DT_GETU32 (pData + 0x18);
     while (dwOffset < dwSize) {
          dwType = DT_GETU32 (pData + dwOffset);

          if (FIoDebug)
               fprintf (FIoDebug,
                        "ReadAllAttributes : found type 0x%" PRIx32 "\n",
                        dwType);

          if (dwType == MAXDWORD)       /* FFFFFFFF : end of attribute list */
               break;

          if (!InitAttribute (pEntry, pData + dwOffset))
               return FALSE;

          dwOffset += DT_GETU32 (pData + dwOffset + 0x4);
     }
     return TRUE;
}

/**********************************************************************
 * CheckRecord : check for validity and apply update sequence to a
 *               (supposed to be) MFT record buffer
 *
 * - [in/out] pData : pointer to buffer containing MFT record
 * - [in] dwDataSize : size of the buffer
 * - [in] dwSectorSize : disk sector size (used by update sequence)
 *
 * returns TRUE in case of success, FALSE otherwise
 *
 *********************************************************************/
int
CheckRecord (unsigned char *pData, uint32_t dwDataSize, uint32_t dwSectorSize)
{
     uint32_t dwArrayOffset, dwWordOffset;
     uint16_t wUpdate, wNumber;
     uint16_t *pTest = NULL;

     if (FIoDebug && BIoDump)
          DumpHexBytes (FIoDebug, pData, dwDataSize);

     if (!IS_MFT_RECORD (pData)) {
          ALARM (TA_PROG_ERR, "not a file record");
          return FALSE;
     }
     dwArrayOffset = DT_GETU16 (pData + 0x4);
     wNumber = DT_GETU16 (pData + 0x6);
     /* It's not usefull to correct byte order since we compare values
      * within the record */
     wUpdate = *((uint16_t *) (pData + dwArrayOffset));
     dwArrayOffset += sizeof (uint16_t);
     wNumber--;
     dwWordOffset = dwSectorSize - sizeof (uint16_t);

     while (wNumber) {
          if (dwWordOffset > dwDataSize) {
               ALARM (TA_PROG_ERR, "inconsistent number of update values - "
                      "Record size : %" PRIu32 " # of update values %" PRIu16,
                      dwDataSize, DT_GETU16 (pData + 0x6) - 1);
               return FALSE;
          }
          pTest = (uint16_t *) (pData + dwWordOffset);
          if (wUpdate != *pTest) {
               ALARM (TA_PROG_ERR, "bad sector - "
                      "Update value : %" PRIu16 " sector value : %" PRIu16,
                      wUpdate, *pTest);
               return FALSE;
          }
          *pTest = *((uint16_t *) (pData + dwArrayOffset));
          dwArrayOffset += sizeof (uint16_t);
          wNumber--;
          dwWordOffset += dwSectorSize;

     }

     return TRUE;
}

/**********************************************************************
 * ReadRecord : add to an entry structure the informations of an MFT
 *              record
 *
 * - [in/out] pEntry : pointer to entry (file description) structure
 * - [in] ulNum : index in the MFT of the record to add
 * - [in] bBase : look for a base record if TRUE, an extended one otherwise
 * - [in] bErasedOnly : restrict process to erased records if TRUE
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak, SIMPLE_ERROR otherwise
 *
 *********************************************************************/
int
ReadRecord (PNTFS_MFT_ENTRY pEntry, uint64_t ulNum,
            int bBase, int bErasedOnly)
{
     int iErrorLevel = OK_ERROR;
     unsigned char *pBuffer = NULL, *pStart = NULL;
     uint64_t ulVcnLength;
     uint64_t ulBase;
     uint16_t wSeqNum = 0;

     /* Safe defaults */
     if (bBase) {
          pEntry->wAttrCount = 0;
          pEntry->pAttr = NULL;
     }

     if (FIoDebug)
          fprintf (FIoDebug,
                   (bBase ?
                    "\nLooking for base record #%" PRIu64 "\n" :
                    "Looking for extended record #%" PRIu64 "\n"), ulNum);

     pBuffer =
          DT_CALLOC (unsigned char,
                     pEntry->pVolume->dwMftRecordClusterSize *
                     pEntry->pVolume->dwClusterSize,
                     "pBuffer - Record #%" PRIu64, ulNum);

/* Since a record can be smaller than a cluster,
 * we have to read the whole cluster and add an offset */
     pStart = pBuffer +
          ((pEntry->pVolume->dwMftRecordSize * ulNum) %
           pEntry->pVolume->dwClusterSize);
     ulVcnLength = (uint64_t) pEntry->pVolume->dwMftRecordClusterSize;
     if (!ReadStreamVcns (pEntry->pVolume,
                          pEntry->pVolume->pMftData,
                          (pEntry->pVolume->dwMftRecordSize * ulNum) /
                          pEntry->pVolume->dwClusterSize, &ulVcnLength,
                          pBuffer)) {
          ALARM (TA_PROG_ERR, "ReadStreamVcns failure - Record #%" PRIu64,
                 ulNum);
          iErrorLevel = SIMPLE_ERROR;
     } else if (!CheckRecord (pStart,
                              pEntry->pVolume->dwMftRecordSize,
                              pEntry->pVolume->dwSectorSize)) {
          ALARM (TA_PROG_ERR, "CheckRecord failure - Record #%" PRIu64,
                 ulNum);
          iErrorLevel = SIMPLE_ERROR;
     } else {

          ulBase = DT_GETU48 (pStart + 0x20);
          wSeqNum = DT_GETU16 (pStart + 0x26);

          if ((bBase && (!ulBase) && (!wSeqNum)) ||
              ((!bBase) && (ulBase == pEntry->ulNumber) &&
               (wSeqNum == pEntry->wSequence))) {
               if (bBase) {
                    /* This is a new entry => init phase */
                    pEntry->ulNumber = ulNum;
                    pEntry->wSequence = DT_GETU16 (pStart + 0x10);
                    pEntry->wRefCount = DT_GETU16 (pStart + 0x12);
                    pEntry->wFlags = DT_GETU16 (pStart + 0x16);
                    if (!(pEntry->wFlags & 1))  /* Test the 'in use' bit */
                         pEntry->wRefCount = 0;
                    pEntry->bDirectory =
                         ((pEntry->wFlags & 2) ? TRUE : FALSE);
                    /* Safe defaults */
                    pEntry->wAttrCount = 0;
                    pEntry->pAttr = NULL;
               }

               if (!(pEntry->wRefCount && bErasedOnly))
                    /* No process of not erased records if bErasedOnly is set */
                    if (!ReadAllAttributes (pEntry, pStart))
                         iErrorLevel = SIMPLE_ERROR;
          } else {              /* if ((bBase && (! ulBase)) || ((!bBase)... */

               if (FIoDebug)
                    fprintf (FIoDebug, "Extended record detected\n");
               if (bBase)
                    ALARM (TA_PROG_WARN, "not a base record (base #%" PRIu64
                           ") - Requested record #%" PRIu64, ulBase, ulNum);
               else
                    ALARM (TA_PROG_ERR, "wrong parent file reference %" PRIu64
                           "-%" PRIu16 " (%" PRIu64 "-%" PRIu16
                           " expected) - " "Requested record #%" PRIu64,
                           ulBase, wSeqNum, pEntry->ulNumber,
                           pEntry->wSequence, ulNum);
               iErrorLevel = SIMPLE_ERROR;
          }                     /* if ((bBase && (! ulBase)) || ((!bBase)... */
     }

     DT_FREE (pBuffer, "pBuffer - Requested record #%" PRIu64, ulNum);

     return iErrorLevel;
}

/**********************************************************************
 * ProcessAttributeList : process the stream of the attribute "Attribute
 *                        List" to add extended record informations to
 *                        the given entry
 *
 * - [in/out] pEntry : pointer to entry (file description) structure
 * - [in] pData : pointer to stream buffer (NTFS byte-order)
 * - [in] ulSize : size of the stream
 * - [in] bErasedOnly : restrict process to erased records if TRUE
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak, SIMPLE_ERROR otherwise
 *
 *********************************************************************/
int
ProcessAttributeList (PNTFS_MFT_ENTRY pEntry, unsigned char *pData,
                      uint64_t ulSize, int bErasedOnly)
{
     int iErrorLevel = OK_ERROR;
     uint64_t ulRecord, ulPos = 0;
     uint32_t dwBoucle, dwNum = 0;
     uint64_t *pRecords = NULL, *pNew = NULL;
     int bFound;

     /* Create a buffer to memorize already processed records */
     pRecords = DT_CALLOC (uint64_t, ATTRIBUTE_NUMBER_ALLOCATED,
                           "pRecords - Base record #%" PRIu64,
                           pEntry->ulNumber);

     pRecords[0] = pEntry->ulNumber;
     dwNum++;

     while (ulPos < ulSize) {
          ulRecord = DT_GETU48 (pData + ulPos + 0x10);

          /* Has the record indexed by ulRecord already been processed ? */
          bFound = FALSE;
          for (dwBoucle = 0; dwBoucle < dwNum; dwBoucle++) {
               if (ulRecord == pRecords[dwBoucle]) {
                    bFound = TRUE;
                    break;
               }
          }
          if (!bFound) {
               /* A new extended record has been found
                * Allocate some memory for the index value if needed */
               if (!(dwNum % ATTRIBUTE_NUMBER_ALLOCATED)) {
                    /* Re-allocate memory if buffer for record numbers
                     * is too small */
                    pNew = DT_REALLOC (pRecords, uint64_t, dwNum +
                                       ATTRIBUTE_NUMBER_ALLOCATED,
                                       "Base record #%" PRIu64,
                                       pEntry->ulNumber);
               }

               /*  if (!(dwNum % ATTRIBUTE_NUMBER_ALLOCATED)) */
               /* Put in the buffer the index of the new extended record */
               pRecords[dwNum] = ulRecord;
               dwNum++;

               switch (ReadRecord (pEntry, ulRecord, FALSE, bErasedOnly)) {
               case SIMPLE_ERROR:
                    iErrorLevel = SIMPLE_ERROR;
                    break;
               case FATAL_ERROR:
                    return FATAL_ERROR;
                    break;
               }                /*  switch (ReadRecord(pEntry, ulRecord,... */
               if (iErrorLevel != OK_ERROR)
                    break;
          }                     /* if (!bFound) */
          ulPos += DT_GETU16 (pData + ulPos + 0x4);
     }

     /* Free buffer memory */
     DT_FREE (pRecords, "Base Record #%" PRIu64, pEntry->ulNumber);

     return iErrorLevel;
}

/**********************************************************************
 * FreeAttribute : frees the buffer used by an attribute structure
 *
 * - [in] pAttr : pointer to the attribute structure
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak
 *
 *********************************************************************/
int
FreeAttribute (PNTFS_ATTRIBUTE pAttr)
{
     if (pAttr) {
          DT_FREE (pAttr->pStream, "pStream - Attribute type 0x%" PRIx32,
                   pAttr->dwType);

          DT_FREE (pAttr->pRunlist, "pRunlist -Attribute type 0x%" PRIx32,
                   pAttr->dwType);
          pAttr->ulRunlistSize = 0;

     }
     /* if (pAttr) */
     return OK_ERROR;
}

/**********************************************************************
 * CleanEntry : frees the buffer used by an entry structure
 *
 * - [in] pEntry : pointer to the entry structure
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak
 *
 *********************************************************************/
int
CleanEntry (PNTFS_MFT_ENTRY pEntry)
{
     uint16_t wCount;

     if (FIoDebug)
          fprintf (FIoDebug,
                   "Cleaning entry, base record #%" PRIu64 "\n",
                   pEntry->ulNumber);

     if (pEntry->pAttr) {
          if (pEntry->wAttrCount) {
               for (wCount = 0; wCount < pEntry->wAttrCount; wCount++) {
                    if (FreeAttribute (&pEntry->pAttr[wCount]) == FATAL_ERROR)
                         return FATAL_ERROR;
               }                /* for (wCount = 0; wCount < pEntry... */
          }                     /* if (pEntry->wAttrCount) */
          DT_FREE (pEntry->pAttr, "pAttr - Record #%" PRIu64,
                   pEntry->ulNumber);
          pEntry->wAttrCount = 0;

     }                          /* if (pEntry->pAttr) */
     pEntry->ulNumber = 0;

     return OK_ERROR;
}

/**********************************************************************
 * FreeEntry : frees the buffer used by an entry structure including
 *              itself
 *
 * - [in] pEntry : pointer to the entry structure
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak
 *
 *********************************************************************/
int
FreeEntry (PNTFS_MFT_ENTRY pEntry)
{
     if (pEntry) {
          if (CleanEntry (pEntry) == FATAL_ERROR)
               return FATAL_ERROR;

          DT_FREE (pEntry, "pEntry");

     }                          /* if (pEntry) */
     return OK_ERROR;
}

/**********************************************************************
 * ReadEntry : fill an entry structure according to the base record
 *             index given
 *
 * - [in/out] pEntry : pointer to entry (file description) structure
 * - [in] ulNum : index in the MFT of the file base record
 * - [in] bErasedOnly : restrict process to erased records if TRUE
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak, SIMPLE_ERROR otherwise
 *
 *********************************************************************/
int
ReadEntry (PNTFS_MFT_ENTRY pEntry, uint64_t ulNum, int bErasedOnly)
{
     int iErrorLevel = OK_ERROR;
     PNTFS_ATTRIBUTE pList = NULL;
     unsigned char *pBuffer = NULL;
     uint64_t ulSize;

     pEntry->dwSecId = 0;
     pEntry->bSecIdFound = FALSE;
     switch (ReadRecord (pEntry, ulNum, TRUE, bErasedOnly)) {
     case SIMPLE_ERROR:
          if (CleanEntry (pEntry) == FATAL_ERROR)
               return FATAL_ERROR;
          return SIMPLE_ERROR;
          break;
     case FATAL_ERROR:
          return FATAL_ERROR;
          break;
     }

     /* Looking for an attribute list */
     pList = FindAttribute (pEntry, pEntry->pVolume->dwAtAttrList, NULL);
     if (pList && (!(pEntry->wRefCount && bErasedOnly))) {
          /* AttributeList$ found
             Do not process not erased records if bErasedOnly is set */
/* pBuffer = GetStreamBuffer(pEntry, pList, &ulSize); */
          pBuffer = DT_CALLOC (unsigned char, pList->ulStreamSize,
                               "pBuffer - Base record #%" PRIu64, ulNum);

          ulSize = pList->ulStreamSize;
          if (ReadStreamBytes (pEntry->pVolume, pList, 0, &ulSize, pBuffer)) {
/* Tester la valeur de retour */
               ProcessAttributeList (pEntry,
                                     pBuffer,
                                     pList->ulStreamSize, bErasedOnly);
          } else {              /* if (ReadStreamBytes(pVol, pList... */

               ALARM (TA_PROG_ERR,
                      "ReadStreamBytes failure - Base record #%" PRIu64,
                      ulNum);
               iErrorLevel = SIMPLE_ERROR;
          }

/* ReleaseStreamBuffer(pList, pBuffer) */
          DT_FREE (pBuffer, "pBuffer - Base record #%" PRIu64, ulNum);
     }

     if (FIoDebug)
          DumpEntry (FIoDebug, pEntry);

     return iErrorLevel;
}

/**********************************************************************
 * InitEntry : allocate and fill an entry structure according to the
 *             base record index given
 *
 * - [in] pVol : pointer to volume description structure
 * - [out] pEntry : pointer to pointer to initialize
 * - [in] ulNum : index in the MFT of the file base record
 * - [in] bErasedOnly : restrict process to erased records if TRUE
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak, SIMPLE_ERROR otherwise
 *
 *********************************************************************/
int
InitEntry (PNTFS_VOL_INFO pVol, PNTFS_MFT_ENTRY * pEntry,
           uint64_t ulNum, int bErasedOnly)
{
     *pEntry = NULL;
     *pEntry =
          DT_CALLOC (NTFS_MFT_ENTRY, 1, "*pEntry - Base record #%" PRIu64,
                     ulNum);
     (*pEntry)->pVolume = pVol;

     switch (ReadEntry (*pEntry, ulNum, bErasedOnly)) {
     case SIMPLE_ERROR:
          return SIMPLE_ERROR;
          break;
     case FATAL_ERROR:
          return FATAL_ERROR;
          break;
     }

     return OK_ERROR;
}

/**********************************************************************
 * CloseVolume : frees the buffer used a volume information structure,
 *               close the raw disk access
 *
 * - [in/out] pVol : pointer to volume description structure
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak
 *
 *********************************************************************/
int
CloseVolume (PNTFS_VOL_INFO pVol)
{
#ifdef _WIN32
     if (pVol->hVolume != INVALID_HANDLE_VALUE) {
          if (!CloseHandle (pVol->hVolume)) {
               ALARM (TA_SYS_FATAL, "CloseHandle");
               return FATAL_ERROR;
          }
          pVol->hVolume = INVALID_HANDLE_VALUE;
     }
#else /* _WIN32 */
     if (pVol->hVolume != -1) {
          if (close (pVol->hVolume) == -1) {
               ALARM (TA_SYS_FATAL, "close");
               return FATAL_ERROR;
          }
          pVol->hVolume = -1;
     }
#endif /* _WIN32 */

#ifdef COMPRESSED_STREAMS_SUPPORT
     DT_FREE (PChunkBufferIn, "PChunkBufferIn");
     DT_FREE (PChunkBufferOut, "PChunkBufferOut");

#endif /* COMPRESSED_STREAMS_SUPPORT */

     DT_FREE (PClusterBuffer, "PClusterBuffer");
     DT_FREE (PProcessStreamBuffer, "PProcessStreamBuffer");


     if (pVol->secList.pSecInfo) {
          DT_FREE (pVol->secList.pSecInfo, "pVol->secList.pSecInfo");
     }

     if (pVol->pMft) {
          if (FreeEntry (pVol->pMft) == FATAL_ERROR) {
               ALARM (TA_PROG_FATAL, "FreeEntry pMft");
               return FATAL_ERROR;
          }
     }
     if (pVol->pSecure) {
          if (FreeEntry (pVol->pSecure) == FATAL_ERROR) {
               ALARM (TA_PROG_FATAL, "FreeEntry pSecure");
               return FATAL_ERROR;
          }
     }

     return OK_ERROR;
}

/**********************************************************************
 * AddSecureSD : 
 *********************************************************************/
int
AddSecureSD (PNTFS_VOL_INFO pVol, uint32_t dwSecId, uint64_t ulOffset,
             uint32_t dwSize)
{
#define SDS_INFO_ALLOCATED 20
     PNTFS_SEC_INFO pCopy;

     if (pVol->secList.pSecInfo) {
          if (pVol->secList.dwListSize % SDS_INFO_ALLOCATED) {
               pCopy = DT_REALLOC (pVol->secList.pSecInfo, NTFS_SEC_INFO,
                                   pVol->secList.dwListSize +
                                   SDS_INFO_ALLOCATED, "pCopy");

               pVol->secList.pSecInfo = pCopy;
          }
     } else {
          pVol->secList.pSecInfo =
               DT_CALLOC (NTFS_SEC_INFO, SDS_INFO_ALLOCATED,
                          "pVol->secList.pSecInfo");
          pVol->secList.dwFirst = dwSecId;
          pVol->secList.dwListSize = 0;
          pVol->secList.dwMaxSize = 0;
     }
     pVol->secList.pSecInfo[pVol->secList.dwListSize].dwSecId = dwSecId;
     pVol->secList.pSecInfo[pVol->secList.dwListSize].dwSize = dwSize;
     pVol->secList.pSecInfo[pVol->secList.dwListSize].ulSDSOffset =
          ulOffset + 0x14;
     pVol->secList.dwListSize++;
     pVol->secList.dwMaxSize = max (dwSize, pVol->secList.dwMaxSize);
     if (FIoDebug)
          fprintf (FIoDebug, "Added to pVol->secList : Id %" PRIu32
                   ", size %" PRIu32 ", offset %" PRIu64 "\n",
                   dwSecId, dwSize, ulOffset);
     return OK_ERROR;
}

/**********************************************************************
 * ParseSecureSDS : 
 *********************************************************************/
int
ParseSecureSDS (PNTFS_VOL_INFO pVol)
{
#define SDS_BUFFER_SIZE 0x20

     uint64_t ulOffset = 0, ulSize = SDS_BUFFER_SIZE, ulRead = 0;
     uint32_t dwSecId, dwSize, dwLeft;
     unsigned char Buffer[SDS_BUFFER_SIZE];

     while (ulOffset < pVol->pSecureData->ulStreamSize) {
          ulSize =
               min (SDS_BUFFER_SIZE,
                    pVol->pSecureData->ulStreamSize - ulOffset);
          ReadStreamBytes (pVol, pVol->pSecureData, ulOffset, &ulSize,
                           Buffer);
          if (ulSize >= 0x14) {
               if ((ulRead = DT_GETU64 (Buffer + 8)) == ulOffset &&
                   (dwSize = DT_GETU32 (Buffer + 0x10)) >= 0x14) {
                    dwSecId = DT_GETU32 (Buffer + 4);
                    dwLeft = dwSize % 0x10;
                    AddSecureSD (pVol, dwSecId, ulOffset, dwSize);
                    ulOffset += (dwSize - dwLeft) + (dwLeft ? 0x10 : 0);
               } else {
/*                    break; */
                    /* 16 bytes boundary alignement */
                    ulOffset += 0x10;
               }
          } else {
               break;
          }
     }
     return OK_ERROR;
}

/**********************************************************************
 * OpenVolume : initialize a volume information structure, open in read
 *              only mode the raw disk
 *
 * - [in/out] pVol : pointer to volume description structure
 *
 * returns OK_ERROR in case of success, FATAL_ERROR in case of memory
 *         leak, SIMPLE_ERROR otherwise
 *
 *********************************************************************/
int
OpenVolume (PNTFS_VOL_INFO pVol)
{
/* Improvement : use $MftMirr in case of failures... */
     int iErrorLevel = OK_ERROR;
#ifdef _WIN32
     char cVolName[] = "\\\\.\\A:";
     char cRootPath[] = "A:\\";
#endif /* _WIN32 */
     NTFS_MFT_ENTRY eMft;
     unsigned char *pBuffer = NULL;
     uint64_t ulSize;
     PNTFS_MFT_ENTRY pEntry = NULL, pTempEntry = NULL;
     PNTFS_ATTRIBUTE pVolInfo = NULL;
#ifdef _WIN32
     uint32_t dwFreeClusters, dwTotalClusters;
#endif /* _WIN32 */
     uint32_t dwClusterSize, dwClusterSectorSize, dwSectorSize;
     signed char cRecordSize = 0;

     PProcessStreamBuffer = NULL;
#ifdef COMPRESSED_STREAMS_SUPPORT
     PChunkBufferIn = NULL;
     PChunkBufferOut = NULL;
#endif /* COMPRESSED_STREAMS_SUPPORT */
     PClusterBuffer = NULL;

     pVol->pMft = NULL;
     pVol->pSecure = NULL;
     pVol->pMftData = NULL;
     pVol->pSecureData = NULL;
     pVol->secList.dwListSize = 0;
     pVol->secList.dwFirst = 0;
     pVol->secList.dwMaxSize = 0;
     pVol->secList.pSecInfo = NULL;

#ifdef _WIN32
     cRootPath[0] = pVol->tDrive;
     if (!GetDiskFreeSpace (cRootPath,
                            &dwClusterSectorSize,
                            &dwSectorSize,
                            &dwFreeClusters, &dwTotalClusters)) {
          ALARM (TA_SYS_ERR, "GetDiskFreeSpace - Path : %s", cRootPath);
          return SIMPLE_ERROR;
     }
#else /* _WIN32 */
     dwSectorSize = 512;
     dwClusterSectorSize = 1;
#endif /* _WIN32 */
     dwClusterSize = dwSectorSize * dwClusterSectorSize;
     pVol->dwClusterSize = dwClusterSize;
     pVol->ulClusterNumber = 1; /* Fake value to read the boot cluster */
     DwMaxRead = ((uint32_t) (MAXDWORD / pVol->dwClusterSize)) *
          pVol->dwClusterSize;

#ifdef _WIN32
     cVolName[4] = pVol->tDrive;
     pVol->hVolume = CreateFile (cVolName,
                                 GENERIC_READ,
                                 FILE_SHARE_READ | FILE_SHARE_WRITE,
                                 NULL, OPEN_EXISTING, 0, 0);
     if (pVol->hVolume == INVALID_HANDLE_VALUE) {
          ALARM (TA_SYS_ERR, "CreateFile - Drive : %s", cVolName);
          return SIMPLE_ERROR;
     }
#else /* _WIN32 */
     if ((pVol->hVolume = open (pVol->file, O_RDONLY)) == -1) {
          ALARM (TA_SYS_ERR, "open - File : %s", pVol->file);
          return SIMPLE_ERROR;
     }
#endif /* _WIN32 */

/* Buffer creation to get the boot cluster */
     ulSize = 1;
#ifdef _WIN32
     PClusterBuffer = DT_CALLOC (unsigned char, dwClusterSize,
                                 "PClusterBuffer - Drive : %s", cVolName);
#else
     PClusterBuffer = DT_CALLOC (unsigned char, dwClusterSize,
                                 "PClusterBuffer - File : %s", pVol->file);
#endif
     if (!ReadClusters (pVol, 0, &ulSize, PClusterBuffer)) {
          iErrorLevel = SIMPLE_ERROR;
     } else if (!IS_NTFS_VOL (PClusterBuffer + 3)) {
#ifdef _WIN32
          ALARM (TA_PROG_ERR, "drive %s is not NTFS", cVolName);
#else /* _WIN32 */
          ALARM (TA_PROG_ERR, "file %s is not NTFS partition/image",
                 pVol->file);
#endif /* _WIN32 */
          if (FIoDebug && BIoDump)
               DumpHexBytes (FIoDebug, PClusterBuffer, dwClusterSize);
          iErrorLevel = SIMPLE_ERROR;
     } else {
          if (FIoDebug && BIoDump)
               DumpHexBytes (FIoDebug, PClusterBuffer, dwClusterSize);
          pVol->dwSectorSize = (uint32_t) DT_GETU16 (PClusterBuffer + 0xB);
          dwClusterSectorSize = (uint32_t) DT_GETU8 (PClusterBuffer + 0xD);
          pVol->dwClusterSize = pVol->dwSectorSize * dwClusterSectorSize;
          pVol->ulClusterNumber =
               DT_GETU64 (PClusterBuffer + 0x28) / dwClusterSectorSize;
          pVol->ulMftStart = DT_GETU64 (PClusterBuffer + 0x30);
          pVol->ulMftMirrStart = DT_GETU64 (PClusterBuffer + 0x38);
/*      pVol->dwMftRecordClusterSize = DT_GETU32(pBuffer + 0x40);
      if (pVol->dwMftRecordClusterSize == 0xF6)
        {
          pVol->dwMftRecordSize = pVol->dwClusterSize / 4;
          pVol->dwMftRecordClusterSize = 1; 
        }
      else
        pVol->dwMftRecordSize = pVol->dwMftRecordClusterSize *
                                pVol->dwClusterSize; */
          cRecordSize = DT_GETS8 (PClusterBuffer + 0x40);
          if (cRecordSize < 0) {
               cRecordSize = -cRecordSize;
               pVol->dwMftRecordSize = 1 << cRecordSize;
               pVol->dwMftRecordClusterSize = (pVol->dwMftRecordSize +
                                               pVol->dwClusterSize - 1) /
                    pVol->dwClusterSize;
          } else {
               pVol->dwMftRecordClusterSize = (uint32_t) cRecordSize;
               pVol->dwMftRecordSize = pVol->dwMftRecordClusterSize *
                    pVol->dwClusterSize;
          }
          pVol->dwIndexRecordClusterSize = DT_GETU32 (PClusterBuffer + 0x44);
#ifdef _WIN32
          if ((pVol->dwClusterSize != dwClusterSize) ||
              (pVol->dwSectorSize != dwSectorSize)) {
               ALARM (TA_PROG_ERR,
                      "incoherent values\nValue\tSystem\t"
                      "Boot Sector\nCluster Size\t%" PRIu32 "\t%" PRIu32
                      "\nSectorSize\t%" PRIu32 "\t%" PRIu32,
                      dwClusterSize,
                      pVol->dwClusterSize, dwSectorSize, pVol->dwSectorSize);
               iErrorLevel = SIMPLE_ERROR;
          }
#endif /* _WIN32 */
     }

     if (FIoDebug)
          fprintf (FIoDebug,
                   "SectorSize: 0x%" PRIx32 ", ClusterSize: 0x%" PRIx32
                   ", MftRecordSize: 0x%" PRIx32 "\nClusterNumber: 0x%" PRIx64
                   ", MftStart: 0x%" PRIx64 "\n",
                   pVol->dwSectorSize,
                   pVol->dwClusterSize,
                   pVol->dwMftRecordSize,
                   pVol->ulClusterNumber, pVol->ulMftStart);

     /* Reallocate PClusterBuffer to the real size of dwClusterSize */
     pBuffer = DT_REALLOC (PClusterBuffer, unsigned char, pVol->dwClusterSize,
                           "PClusterBuffer");
     PClusterBuffer = pBuffer;

     if (iErrorLevel != OK_ERROR)
          return iErrorLevel;

     pVol->dwChunkSize = pVol->dwClusterSize * CHUNK_CLUSTER_SIZE;
     /* Define buffer size about 1 Mo and multiple of ChunkSize
      * used in ProcessStream */
     DwProcessStreamBuffer = ((uint32_t) (0x1000000 / pVol->dwChunkSize)) *
          pVol->dwChunkSize;

     /* Default values for attribute types */
     pVol->dwAtStdInfo = 0x10;
     pVol->dwAtAttrList = 0x20;
     pVol->dwAtFileName = 0x30;
     pVol->dwAtObjectID = 0x40; /* Used to be VolVersion */
     pVol->dwAtSecurity = 0x50;
     pVol->dwAtVolName = 0x60;
     pVol->dwAtVolInfo = 0x70;
     pVol->dwAtData = 0x80;
     pVol->dwAtIndexRoot = 0x90;
     pVol->dwAtIndexAlloc = 0xA0;
     pVol->dwAtBitmap = 0xB0;
     pVol->dwAtReparsePoint = 0xC0;     /* Used to be Symlink */
     pVol->dwAtEaInf = 0xD0;
     pVol->dwAtEa = 0xE0;
     pVol->dwAtPropSet = 0xF0;
     pVol->dwAtLogStream = 0x100;

#ifdef COMPRESSED_STREAMS_SUPPORT
     /* Create a buffer to read chunks (16 clusters) used in compression */
     PChunkBufferIn = DT_CALLOC (unsigned char, pVol->dwChunkSize,
                                 "PChunkBufferIn");

     /* Create a buffer to output chunks (16 clusters) used in compression */
     PChunkBufferOut = DT_CALLOC (unsigned char, pVol->dwChunkSize,
                                  "PChunkBufferOut");

#endif /* COMPRESSED_STREAMS_SUPPORT */

     /* eMft is an entry corresponding to the $MFT base record
      * We can't use ReadEntry (calling ReadRecord) since $MFT information
      * is not present yet */
     eMft.pVolume = pVol;

     ulSize = pVol->dwMftRecordClusterSize;
     pBuffer = NULL;
#ifdef _WIN32
     pBuffer = DT_CALLOC (unsigned char, pVol->dwMftRecordClusterSize *
                          pVol->dwClusterSize, "pBuffer - Drive : %s",
                          cVolName);
#else /* _WIN32 */
     pBuffer = DT_CALLOC (unsigned char, pVol->dwMftRecordClusterSize *
                          pVol->dwClusterSize, "pBuffer - File : %s",
                          pVol->file);
#endif /* _WIN32 */
     if (!ReadClusters (pVol, pVol->ulMftStart, &ulSize, pBuffer)) {
          iErrorLevel = SIMPLE_ERROR;
     } else if (!CheckRecord (pBuffer,
                              pVol->dwMftRecordSize, pVol->dwSectorSize)) {
          iErrorLevel = SIMPLE_ERROR;
     } else {
          eMft.ulNumber = 0;
          eMft.wSequence = DT_GETU16 (pBuffer + 0x10);
          eMft.wRefCount = DT_GETU16 (pBuffer + 0x12);
          eMft.wFlags = DT_GETU16 (pBuffer + 0x16);
          /* Safe defaults */
          eMft.wAttrCount = 0;
          eMft.pAttr = NULL;
/* FIXME check returned value */
          ReadAllAttributes (&eMft, pBuffer);
          pVol->pMft = &eMft;
          pVol->pMftData = FindAttribute (&eMft, pVol->dwAtData, NULL);
          if (!pVol->pMftData) {
#ifdef _WIN32
               ALARM (TA_PROG_ERR,
                      "No Data Attribute in $MFT base record - Drive : %s",
                      cVolName);
#else /* _WIN32 */
               ALARM (TA_PROG_ERR,
                      "No Data Attribute in $MFT base record - File : %s",
                      pVol->file);
#endif /* _WIN32 */
               iErrorLevel = SIMPLE_ERROR;
          }
     }

     if (FIoDebug)
          DumpEntry (FIoDebug, &eMft);

     /* Free the read buffer */
#ifdef _WIN32
     DT_FREE (pBuffer, "pBuffer - Drive : %s", cVolName);
#else
     DT_FREE (pBuffer, "pBuffer - File : %s", pVol->file);
#endif

     if (iErrorLevel != OK_ERROR)
          return iErrorLevel;

/* Should be useless to correct with $AttrDef since the default
 * attributes definition has already been used to get here...   
 * We get $AttrDef informations to correct (??) the values
 * used for the attributes types
  switch (InitEntry(pVol, &pEntry, 0x4, FALSE))
    {
      case SIMPLE_ERROR:
        return SIMPLE_ERROR;
        break;
      case FATAL_ERROR:
        return FATAL_ERROR;
        break;
    }
 * TO BE COMPLETED
 * Read the $DATA Stream and process it
 * pBuffer = GetStreamBuffer(pEntry, pAttr, &ulLength);
 * ProcessAttrDef(pVol, pBuffer);
 * ReleaseStreamBuffer(pAttr, pBuffer);

  if (CleanEntry(pEntry) == FATAL_ERROR)
    return FATAL_ERROR;
*/

     /* Now we can use the information in the base $MFT record
      * to get the whole $MFT entry including extended records */

     switch (InitEntry (pVol, &pEntry, 0, FALSE)) {
     case SIMPLE_ERROR:
          return SIMPLE_ERROR;
          break;
     case FATAL_ERROR:
          return FATAL_ERROR;
          break;
     }
     pVol->pMft = pEntry;
     pVol->pMftData = FindAttribute (pEntry, pVol->dwAtData, NULL);
     if (!pVol->pMftData) {
#ifdef _WIN32
          ALARM (TA_PROG_ERR,
                 "No Data Attribute in $MFT entry - Drive : %s", cVolName);
#else /* _WIN32 */
          ALARM (TA_PROG_ERR,
                 "No Data Attribute in $MFT entry - File : %s", pVol->file);
#endif /* _WIN32 */
          return SIMPLE_ERROR;
     }
     /* Free attributes in eMft */
     if (CleanEntry (&eMft) == FATAL_ERROR)
          return FATAL_ERROR;

     pVol->wMajorVersion = 1;   /*NT4 safe default */
     if (InitEntry (pVol, &pTempEntry, 3, FALSE) == OK_ERROR) {
          pVolInfo = FindAttribute (pTempEntry, pVol->dwAtVolInfo, NULL);
          if (pVolInfo) {
               pBuffer = GetStreamBuffer (pTempEntry, pVolInfo, &ulSize);
               if (ulSize >= 8)
                    pVol->wMajorVersion = DT_GETU8 (pBuffer + 8);
               else
                    ALARM (TA_PROG_ERR,
                           "InitVolume : Volume Information truncated stream - "
                           "Size : %" PRIu64, ulSize);

               if (ReleaseStreamBuffer (pVolInfo, pBuffer) == FATAL_ERROR)
                    return FATAL_ERROR;
          } else {
#ifdef _WIN32
               ALARM (TA_PROG_ERR,
                      "InitVolume : attribute 0x%" PRIx32
                      " from base record 3 not found, setting MajorVersion=1 - "
                      "Drive : %s", pVol->dwAtVolInfo, cVolName);
#else /* _WIN32 */
               ALARM (TA_PROG_ERR,
                      "InitVolume : attribute 0x%" PRIx32
                      " from base record 3 not found, setting MajorVersion=1 - "
                      "File : %s", pVol->dwAtVolInfo, pVol->file);
#endif /* _WIN32 */
          }
     } else {
#ifdef _WIN32
          ALARM (TA_PROG_ERR,
                 "base record 3 not found - Drive : %s", cVolName);
#else /* _WIN32 */
          ALARM (TA_PROG_ERR,
                 "base record 3 not found - File : %s", pVol->file);
#endif /* _WIN32 */
     }
     if (FIoDebug)
          fprintf (FIoDebug, "Filesystem major version : %" PRIu16 "\n",
                   pVol->wMajorVersion);

     /* Processing Secure SDS attribute */
     if (pVol->wMajorVersion >= 3) {
          if (CleanEntry (pTempEntry) == FATAL_ERROR)
               return FATAL_ERROR;
          pTempEntry->pVolume = pVol;
          if (ReadEntry (pTempEntry, 9, FALSE) == OK_ERROR) {
               pVol->pSecure = pTempEntry;
               pVol->pSecureData =
                    FindAttribute (pTempEntry, pVol->dwAtData, NULL);
               if (pVol->pSecureData)
                    ParseSecureSDS (pVol);
               else {
#ifdef _WIN32
                    ALARM (TA_PROG_ERR,
                           "attribute 0x%" PRIx32
                           " from base record 9 not found - Drive : %s",
                           pVol->dwAtData, cVolName);
#else /* _WIN32 */
                    ALARM (TA_PROG_ERR,
                           "attribute 0x%" PRIx32
                           " from base record 9 not found\nFile : %s",
                           pVol->dwAtVolInfo, pVol->file);
#endif /* _WIN32 */
               }
          } else {
#ifdef _WIN32
               ALARM (TA_PROG_ERR,
                      "base record 9 not found - Drive : %s", cVolName);
#else /* _WIN32 */
               ALARM (TA_PROG_ERR,
                      "base record 9 not found\nFile : %s", pVol->file);
#endif /* _WIN32 */
               pVol->pSecure = NULL;
               pVol->pSecureData = NULL;
               if (pTempEntry)
                    return FreeEntry (pTempEntry);
          }
     } else {
          if (pTempEntry)
               return FreeEntry (pTempEntry);
     }

     return OK_ERROR;
}
