/******************************************************************************
 *
 * sd_dump.c - Security Descriptor buffer dump. Part of the DirTools project.
 *
 * $Revision: 1.1.1.1 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <inttypes.h>
#ifdef _WIN32
#include <windows.h>
#include "winsid.h"
#endif /* _WIN32 */

#include "sd_dump.h"
#include "macros.h"
#include "dt_error.h"
#include "dt_comment.h"
#include "info.h"
#include "dt_memory.h"

TYPE_COMMENT_BITS CommentAccessMask[] = {
     {FILE_ALL_ACCESS, {"Fa", "File all access"}},
     {FILE_GENERIC_READ, {"Fr", "File generic read"}},
     {FILE_GENERIC_WRITE, {"Fw", "File generic write"}},
     {FILE_GENERIC_EXECUTE, {"Fe", "File generic execute"}},
/*   {STANDARD_RIGHTS_REQUIRED, { "Sq", "Standard rights required"}}, */
     {STANDARD_RIGHTS_ALL, {"Sa", "Standard rights all"}},
/*Filesystem specific rights*/
     {FILE_READ_DATA, {"Rd", "File read data"}},
     {FILE_WRITE_DATA, {"Wd", "File write data"}},
     {FILE_APPEND_DATA, {"Ad", "File append data"}},
     {FILE_READ_EA, {"Re", "File read extended area"}},
     /*  ea = extended area, compatibilite avec HPFS (OS2) */   
     {FILE_WRITE_EA, {"We", "File write extended area"}},
     {FILE_EXECUTE, {"Ex", "File execute"}},
     {FILE_DELETE_CHILD, {"Dc", "File delete child"}},
     {FILE_READ_ATTRIBUTES, {"Ra", "File read atributes"}},
     {FILE_WRITE_ATTRIBUTES, {"Wa", "File write attributes"}},
/*Standards rights*/
     {DELETE, {"De", "Delete access"}}, /*  Droit de suppression */
     {READ_CONTROL, {"Rc", "Read control"}},    /* Droit lecture DACL,
                                                 * proprio, groupe */
     {WRITE_DAC, {"Wc", "Write access DACL"}},  /* Droit modifier DACL */
     {WRITE_OWNER, {"Wo", "Write owner"}},      /* Droit de modifier le
                                                 * proprio */
     {SYNCHRONIZE, {"Sy", "Synchronize access"}},       /* Droit de
                                                         * synchronizer ??? */
/*generics  rights*/
     /*  Accs aux SACE */
     {ACCESS_SYSTEM_SECURITY, {"As", "Access system security"}},
     {MAXIMUM_ALLOWED, {"Mx", "Maximum allowed"}},
     {GENERIC_READ, {"Gr", "Generic read"}},    /* Droits en lecture */
     {GENERIC_WRITE, {"Gw", "Generic write"}},  /*  Droits en lecture */
     {GENERIC_EXECUTE, {"Ge", "Generic execute"}},      /* Droits d'xcution */
     {GENERIC_ALL, {"Ga", "Generic all"}},      /* Droit a tout type d'acces */
     {0, {NULL, NULL}}
};

TYPE_COMMENT_BITS CommentAceFlag[] = {
     {OBJECT_INHERIT_ACE, {NULL, "Object Inherit"}},    /* Les non "conteneurs"
                                                         * hritent */
     /*  Seuls les conteneurs (ex : rpertoire) hritent */
     {CONTAINER_INHERIT_ACE, {NULL, "Container Inherit"}},
     /*  Ne plus propager les proprits hrites */
     {NO_PROPAGATE_INHERIT_ACE, {NULL, "No Propagate Inherit"}},
     {INHERIT_ONLY_ACE, {NULL, "Inherit Only"}},        /*  Applicable au(x)
                                                         * fils mais pas 
                                                         * l'objet */
     /*  Audit des succs */
     {SUCCESSFUL_ACCESS_ACE_FLAG, {NULL, "Successfull Access"}},
     {FAILED_ACCESS_ACE_FLAG, {NULL, "Failed Access"}}, /*  Audit des checs */
     {0, {NULL, NULL}}
};

TYPE_COMMENT_BITS CommentControlSDFlag[] = {
     {SE_OWNER_DEFAULTED, {NULL, "Defaulted owner"}},   /* Propritaire par
                                                         * dfaut appliqu */
     {SE_GROUP_DEFAULTED, {NULL, "Defaulted group"}},   /* Groupe par dfaut
                                                         * appliqu */
     {SE_DACL_PRESENT, {NULL, "DACL present"}}, /*  Prsence d'une DACL */
     /*  DACL par dfaut : non gnre par le crateur */
     {SE_DACL_DEFAULTED, {NULL, "Defaulted DACL"}},
     {SE_SACL_PRESENT, {NULL, "SACL present"}}, /*  Prsence d'une SACL */
     /*  SACL par dfaut : non gnre par le crateur */
     {SE_SACL_DEFAULTED, {NULL, "Defaulted SACL"}},
     {SE_DACL_AUTO_INHERIT_REQ, {NULL, "DACL auto inherit req"}},
     {SE_SACL_AUTO_INHERIT_REQ, {NULL, "SACL auto inherit req"}},
     {SE_DACL_AUTO_INHERITED, {NULL, "DACL auto inherited"}},
     {SE_SACL_AUTO_INHERITED, {NULL, "SACL auto inherited"}},
     {SE_SELF_RELATIVE, {NULL, "Self relative SD"}},
     {0, {NULL, NULL}}
};

/**********************************************************************
 * DumpSid
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
#ifdef _WIN32
int
DumpSid (FILE * fOutput, unsigned char *pBuffer, uint64_t ulLength,
         int bCompact)
{
     int iErrorLevel = OK_ERROR;
     unsigned char byRevision, bySubNumber = 0, byCount;
     uint64_t ulAuthority, ulOffset = 0;
     uint32_t dwSubAuth;
#define MAX_SUB_AUTHORITY 8
     uint32_t dwSubTab[MAX_SUB_AUTHORITY];
     SID_IDENTIFIER_AUTHORITY siAuthority;
     PSID pSid = 0;
     uint32_t dwRequiredLength = 0;
     static char *pTextualSid = NULL;   /* Let program exit free this buffer */
     static uint32_t dwTextualLength;
     char *pCopy = NULL, *pComment = NULL, *pType = NULL;


/*  for (byCount = 0; byCount <  MAX_SUB_AUTHORITY; byCount++)
      dwSubTab[byCount] = 0; */


     if (ulLength >= 8) {
          byRevision = DT_GETU8 (pBuffer);
          bySubNumber = DT_GETU8 (pBuffer + 1);
          ulAuthority = 0;
          ulOffset = 2;
          /* Compute buffer size
             S-SID_REVISION- + Authority + -Sub-Authorities + NULL
             S- : 2 char
             SID_REVISION : unsigned char : 3 tchar max
             - : 1 char
             Authority : 6 bytes : 15 char
             -Sub-Authorities : uint32_t + 1 : 1 + 10 char */
          dwRequiredLength = 2 + 3 + 1 + 15 + (11 * bySubNumber) + 1;
          if (!pTextualSid) {
               pTextualSid = DT_CALLOC (unsigned char, dwRequiredLength,
                                        "pTextualSid");
               dwTextualLength = dwRequiredLength;
          } else if (dwRequiredLength > dwTextualLength) {
               pCopy =
                    DT_REALLOC (pTextualSid, unsigned char, dwRequiredLength,
                                "pCopy");
               dwTextualLength = dwRequiredLength;
               pTextualSid = pCopy;
          }
          while (ulOffset < 8) {
               ulAuthority <<= 8;
               byCount = DT_GETU8 (pBuffer + ulOffset);
               ulAuthority += byCount;
               siAuthority.Value[ulOffset - 2] = byCount;
               ulOffset++;
          }
          dwRequiredLength = sprintf (pTextualSid,
                                      "S-%u-%" PRIu64 "",
                                      byRevision, ulAuthority);
     }

     /* Writing Sub-Authorities */
     byCount = 0;
     while (((ulOffset + 4) <= ulLength) && (byCount < bySubNumber)) {
          dwSubAuth = DT_GETU32 (pBuffer + ulOffset);
          dwRequiredLength += sprintf (pTextualSid + dwRequiredLength,
                                       "-%" PRIu32, dwSubAuth);
          ulOffset += 4;
          if (byCount < MAX_SUB_AUTHORITY)
               dwSubTab[byCount] = dwSubAuth;
          byCount++;
     }

     fprintf (fOutput, "%s", pTextualSid);

     if (((!bCompact) || FUserFile) &&
         bySubNumber && (bySubNumber <= MAX_SUB_AUTHORITY)) {
          if (!AllocateAndInitializeSid (&siAuthority, bySubNumber,
                                         dwSubTab[0], dwSubTab[1],
                                         dwSubTab[2], dwSubTab[3],
                                         dwSubTab[4], dwSubTab[5],
                                         dwSubTab[6], dwSubTab[7], &pSid)) {
               ALARM (TA_SYS_ERR, "AllocateAndInitalizeSid");
          } else {
               if (IsValidSid (pSid)) {
/*              if (FInfoDebug)
                  fprintf(FInfoDebug, "DumpSid : valid SID built\n"); */
                    switch (GetSidHash (pSid, pTextualSid, &pComment, &pType)) {
                    case OK_ERROR:
                         if (!bCompact)
                              fprintf (fOutput, " (%s) %s", pType, pComment);
                         break;
                    case FATAL_ERROR:
                         return FATAL_ERROR;
                         break;
                    case SIMPLE_ERROR:
                         iErrorLevel = SIMPLE_ERROR;
                         break;
                    }
               }

               FreeSid (pSid);
          }
     }
     fprintf (fOutput, (bCompact ? "|" : "\n"));

     return iErrorLevel;
}
#else /* _WIN32 */

int
DumpSid (FILE * fOutput, unsigned char *pBuffer, uint64_t ulLength,
         int bCompact)
{
     unsigned char byRevision, bySubNumber = 0, byCount;
     uint64_t ulAuthority, ulOffset = 0;
     uint32_t dwSubAuth;


     if (ulLength >= 8) {
          byRevision = DT_GETU8 (pBuffer);
          bySubNumber = DT_GETU8 (pBuffer + 1);
          ulAuthority = 0;
          ulOffset = 2;
          while (ulOffset < 8) {
               ulAuthority <<= 8;
               byCount = DT_GETU8 (pBuffer + ulOffset);
               ulAuthority += byCount;
               ulOffset++;
          }
          fprintf (fOutput, "S-%u-%" PRIu64, byRevision, ulAuthority);
     }

     /* Writing Sub-Authorities */
     byCount = 0;
     while (((ulOffset + 4) <= ulLength) && (byCount < bySubNumber)) {
          dwSubAuth = DT_GETU32 (pBuffer + ulOffset);
          fprintf (fOutput, "-%" PRIu32, dwSubAuth);
          ulOffset += 4;
          byCount++;
     }

     fprintf (fOutput, (bCompact ? "|" : "\n"));

     return OK_ERROR;
}
#endif /* _WIN32 */

/**********************************************************************
 * DumpAcl
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
int
DumpAcl (FILE * fOutput, unsigned char *pBuffer, uint64_t ulLength,
         int bCompact)
{
     int iErrorLevel = OK_ERROR;
     unsigned char byType, byFlags;
     uint16_t wAclSize, wNumber, wCount, wOffset, wAceSize;
     uint32_t dwMask;

     if (ulLength >= 6) {
          wAclSize = DT_GETU16 (pBuffer + 2);
          wNumber = DT_GETU16 (pBuffer + 4);
          wOffset = 8;
          if (wAclSize <= ulLength) {
               for (wCount = 0; wCount < wNumber; wCount++) {
                    byType = DT_GETU8 (pBuffer + wOffset);
                    if (!bCompact) {
                         if (byType <= 1)
                              fprintf (fOutput, " Dace #%" PRIu16 "\n",
                                       wCount);
                         if (byType == 2)
                              fprintf (fOutput, " Sace #%" PRIu16 "\n",
                                       wCount);

                         fprintf (fOutput, "  User : ");
                    }

                    byFlags = DT_GETU8 (pBuffer + wOffset + 1);
                    wAceSize = DT_GETU16 (pBuffer + wOffset + 2);
                    dwMask = DT_GETU32 (pBuffer + wOffset + 4);

                    switch (DumpSid (fOutput, pBuffer + wOffset + 8,
                                     ulLength - wOffset, bCompact)) {
                    case SIMPLE_ERROR:
                         iErrorLevel = SIMPLE_ERROR;
                         break;
                    case FATAL_ERROR:
                         return FATAL_ERROR;
                         break;
                    }
                    if (!bCompact) {
                         fprintf (fOutput, "  Flags : ");
                         DumpBits (fOutput, (uint32_t) byFlags,
                                   CommentAceFlag, bCompact);
                    }
                    switch (byType) {
                    case 0:
                         fprintf (fOutput, (bCompact ? "+" : "  Allowed "));
                         DumpBits (fOutput, dwMask, CommentAccessMask,
                                   bCompact);
                         break;
                    case 1:
                         fprintf (fOutput, (bCompact ? "-" : "  Denied "));
                         DumpBits (fOutput, dwMask, CommentAccessMask,
                                   bCompact);
                         break;
                    case 2:
                         if (!bCompact)
                              fprintf (fOutput, "Audit ");
                         if (byFlags == SUCCESSFUL_ACCESS_ACE_FLAG)
                              fprintf (fOutput,
                                       (bCompact ? "+" : "success "));
                         if (byFlags == FAILED_ACCESS_ACE_FLAG)
                              fprintf (fOutput,
                                       (bCompact ? "-" : "failure "));
                         DumpBits (fOutput, dwMask, CommentAccessMask,
                                   bCompact);
                         break;
                    default:
                         fprintf (fOutput, (bCompact ? "|" : "\n"));
                         ALARM (TA_PROG_ERR, "unknown type %u", byType);
                         break;
                    }
                    wOffset += wAceSize;
               }
          }
     }
     return iErrorLevel;
}

/**********************************************************************
 * DumpBufferSecurity
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
int
DumpBufferSecurity (FILE * fOutput, uint64_t ulRef,
                    unsigned char *pBuffer, uint64_t ulLength, int bCompact)
{
     int iErrorLevel = OK_ERROR;
     uint32_t dwFlags, dwUserOffset, dwGroupOffset, dwSaclOffset,
          dwDaclOffset;
     uint16_t wNumber;

     if (ulLength >= 0x14) {
/* Vrif  faire pour dwFlags */
          dwFlags = DT_GETU32 (pBuffer);
          dwUserOffset = DT_GETU32 (pBuffer + 0x4);
          dwGroupOffset = DT_GETU32 (pBuffer + 0x8);
          dwSaclOffset = DT_GETU32 (pBuffer + 0xC);
          dwDaclOffset = DT_GETU32 (pBuffer + 0x10);

          if (!bCompact) {
               fprintf (fOutput, "Security descriptor flags : ");
               DumpBits (fOutput, dwFlags, CommentControlSDFlag, bCompact);
          }

          /* User information */
          if (dwUserOffset)
               if (dwUserOffset < ulLength) {
                    if (!bCompact)
                         fprintf (fOutput, "User : ");

                    switch (DumpSid (fOutput, pBuffer + dwUserOffset,
                                     ulLength - dwUserOffset, bCompact)) {
                    case SIMPLE_ERROR:
                         iErrorLevel = SIMPLE_ERROR;
                         break;
                    case FATAL_ERROR:
                         return FATAL_ERROR;
                         break;
                    }
               } else {
                    ALARM (TA_PROG_ERR,
                           "truncated stream, user missing - Size : %" PRIu64
                           " - Base record/KeyNode offset : %" PRIu64,
                           ulLength, ulRef);
                    iErrorLevel = SIMPLE_ERROR;
                    if (bCompact)
                         fprintf (fOutput, "|");
          } else {              /* if (dwUserOffset) */

               ALARM (TA_PROG_ERR,
                      "user not found - Base record/KeyNode offset : %"
                      PRIu64, ulRef);
               iErrorLevel = SIMPLE_ERROR;
               if (bCompact)
                    fprintf (fOutput, "|");
          }                     /* if (dwUserOffset) */

          /* Group information */
          if (dwGroupOffset)
               if (dwGroupOffset < ulLength) {
                    if (!bCompact)
                         fprintf (fOutput, "Group : ");
                    switch (DumpSid (fOutput, pBuffer + dwGroupOffset,
                                     ulLength - dwGroupOffset, bCompact)) {
                    case SIMPLE_ERROR:
                         iErrorLevel = SIMPLE_ERROR;
                         break;
                    case FATAL_ERROR:
                         return FATAL_ERROR;
                         break;
                    }
               } else {
                    ALARM (TA_PROG_ERR,
                           "truncated stream, group missing - Size : %" PRIu64
                           " - Base record/KeyNode offset : %" PRIu64,
                           ulLength, ulRef);
                    iErrorLevel = SIMPLE_ERROR;
                    if (bCompact)
                         fprintf (fOutput, "|");
          } else {              /* if (dwGroupOffset) */

               ALARM (TA_PROG_ERR,
                      "group not found - Base record/KeyNode offset" " : %"
                      PRIu64, ulRef);
               iErrorLevel = SIMPLE_ERROR;
               if (bCompact)
                    fprintf (fOutput, "|");
          }                     /* if (dwGroupOffset) */

          /* Dacl information */
          if (dwDaclOffset)
               if ((dwDaclOffset + 6) <= ulLength) {
                    wNumber = DT_GETU16 (pBuffer + dwDaclOffset + 4);
                    fprintf (fOutput,
                             (bCompact ? "%u|" : "Number of DACE : %u\n"),
                             wNumber);
               } else {
                    ALARM (TA_PROG_ERR,
                           "truncated stream, DACL missing - Size : %" PRIu64
                           " - Base record/KeyNode offset : %" PRIu64,
                           ulLength, ulRef);
                    iErrorLevel = SIMPLE_ERROR;
                    if (bCompact)
                         fprintf (fOutput, "|");
          } else                /* if (dwDaclOffset) */
               fprintf (fOutput, (bCompact ? "N|" : "No DACL\n"));

          /* Sacl information */
          if (dwSaclOffset)
               if ((dwSaclOffset + 6) <= ulLength) {
                    wNumber = DT_GETU16 (pBuffer + dwSaclOffset + 4);
                    fprintf (fOutput,
                             (bCompact ? "%u|" : "Number of SACE : %u\n"),
                             wNumber);
               } else {
                    ALARM (TA_PROG_ERR,
                           "truncated stream, SACL missing - Size : %" PRIu64
                           " - Base record/KeyNode offset : %" PRIu64,
                           ulLength, ulRef);
                    iErrorLevel = SIMPLE_ERROR;
                    if (bCompact)
                         fprintf (fOutput, "|");
          } else                /* if (dwDaclOffset) */
               fprintf (fOutput, (bCompact ? "N|" : "No SACL\n"));

          if (dwDaclOffset && (ulLength > dwDaclOffset))
               switch (DumpAcl (fOutput, pBuffer + dwDaclOffset,
                                ulLength - dwDaclOffset, bCompact)) {
               case SIMPLE_ERROR:
                    iErrorLevel = SIMPLE_ERROR;
                    break;
               case FATAL_ERROR:
                    return FATAL_ERROR;
                    break;
               }

          if (dwSaclOffset && (ulLength > dwSaclOffset))
               switch (DumpAcl (fOutput, pBuffer + dwSaclOffset,
                                ulLength - dwSaclOffset, bCompact)) {
               case SIMPLE_ERROR:
                    iErrorLevel = SIMPLE_ERROR;
                    break;
               case FATAL_ERROR:
                    return FATAL_ERROR;
                    break;
               }

     } /* if (ulLength >= 0x14) */
     else {
          ALARM (TA_PROG_ERR, "truncated stream - Size : %" PRIu64
                 " - Base record/KeyNode offset : %" PRIu64, ulLength, ulRef);
          iErrorLevel = SIMPLE_ERROR;
          if (bCompact)
               fprintf (fOutput, "||||");
     }

     return iErrorLevel;
}


/**********************************************************************
 * sd_help
 *
 *
 *
 *
 *
 *
 *
 *********************************************************************/
     void
sd_help (FILE * fOutput)
{
     fprintf (fOutput, "\n  File/directory/key access :\n");
     fprintf (fOutput, "   - Specific rights :\n");
     fprintf (fOutput,
              "    Rd : Read data (list directory), "
              "Wd : Write data (add file)\n");
     fprintf (fOutput,
              "    Ad : Append data (add subdirectory), "
           "Re : Read extended area\n");
     fprintf (fOutput,
              "    We : Write extended area, Ex : Execute, "
              "Dc : Delete child\n");
     fprintf (fOutput, "    Ra : Read attributes, Wa : Write attributes\n");
     fprintf (fOutput, "   - Standards rights :\n");
     fprintf (fOutput,
              "    De : Delete, Rc : Read control (owner, group, DACL)\n");
     fprintf (fOutput, "    Wc : Write control DACL, Wo : Write owner\n");
     fprintf (fOutput, "    Sy : Synchronise access\n");
     fprintf (fOutput, "   - Other rights :\n");
     fprintf (fOutput,
              "    As : Access system security, Mx : Maximum allowed\n");
     fprintf (fOutput,
              "    Gr : Generic read, Gw : Generic write, "
              "Ge : Generic execute\n");
     fprintf (fOutput, "    Ga : Generic all\n");
     fprintf (fOutput, "   - Sum up :\n");
     fprintf (fOutput, "    (Sq : Standard rights required = DeRcWcWo)\n");
     fprintf (fOutput, "    Sa : Standard rights all = SqSy\n");
     fprintf (fOutput, "    Fr : File generic read = RdReRaRcSy\n");
     fprintf (fOutput, "    Fw : File generic write = WdAdWeWaRcSy\n");
     fprintf (fOutput, "    Fe : File generic execute = ExRaRcSy\n");
     fprintf (fOutput, "    Fa : File all access = RdWdAdReWeExDcRaWaSa\n");
     fprintf (fOutput, "\n  Concerning DACE and SACE :\n");
     fprintf (fOutput, "    N means none found, 0 means present but empty\n");
     fprintf (fOutput, "    + : Allowed for DACE, audit success for SACE\n");
     fprintf (fOutput, "    - : Refused for DACE, audit failure for SACE\n");

}
