/******************************************************************************
 *
 * sysheader.c - Mactime files header generation. Part of the DirTools project.
 *
 * $Revision: 1.2 $
 *
 * Copyright (c) 2001-2004 Olivier Castan
 *
 * This program/include file is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program/include file is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS
 * distribution in the file COPYING); if not, write to the Free Software
 * Foundation,Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * format : indent -br -ce -cdw -nut -i5
 *
 *****************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <time.h>
#ifdef _WIN32
#include <windows.h>
#else
#include <sys/utsname.h>
#endif /* _WIN32 */

#include "dt_error.h"
#include "dt_locale.h"

/**********************************************************************
 * DumpSystemInfo
 *
 * - [in] fOutput : pointer to file descriptor
 *
 *
 *
 * - [in] bCompact : if TRUE prints output using parseable format
 *
 *********************************************************************/
#ifdef _WIN32
void
DumpSystemInfo (FILE * fOutput, const char *pContentName,
                const char *pMoreHeader, int bCompact)
{
     char tComputerBuffer[MAX_COMPUTERNAME_LENGTH + 1];
     uint32_t dwSize = MAX_COMPUTERNAME_LENGTH;
     OSVERSIONINFO ovInfo;
     SYSTEMTIME stTime;

     if (bCompact) {
          fprintf (fOutput, "Content|Host Name|OS|UTC Time|Local Time|Output Locale|");
          fprintf (fOutput, "%s\n", pMoreHeader);
          fprintf (fOutput, "%s|", pContentName);
     }

     if (GetComputerName (tComputerBuffer, &dwSize)) {
          fprintf (fOutput,
                   (bCompact ? "%s|" : "NetBIOS computer name : %s\n"),
                   tComputerBuffer);
     } else {
          ALARM (TA_SYS_ERR, "GetComputerName");
          if (bCompact)
               fprintf (fOutput, "|");
     }

     ovInfo.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
     if (GetVersionEx (&ovInfo)) {
          if (!bCompact)
               fprintf (fOutput, "Operating system : ");

          if ((ovInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) &&
              (ovInfo.dwMinorVersion == 0))
               fprintf (fOutput, "Win 95");
          else if ((ovInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) &&
                   (ovInfo.dwMinorVersion == 10))
               fprintf (fOutput, "Win 98");
          else if ((ovInfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) &&
                   (ovInfo.dwMinorVersion == 90))
               fprintf (fOutput, "Win Me");
          else if ((ovInfo.dwPlatformId == VER_PLATFORM_WIN32_NT) &&
                   (ovInfo.dwMinorVersion == 51))
               fprintf (fOutput, "Win NT 3.51");
          else if ((ovInfo.dwPlatformId == VER_PLATFORM_WIN32_NT) &&
                   (ovInfo.dwMajorVersion == 4))
               fprintf (fOutput, "Win NT 4.0");
          else if ((ovInfo.dwPlatformId == VER_PLATFORM_WIN32_NT) &&
                   (ovInfo.dwMajorVersion == 5)) {
               if (ovInfo.dwMinorVersion == 0)
                    fprintf (fOutput, "Win 2000");
               else if (ovInfo.dwMinorVersion == 1)
                    fprintf (fOutput, "Win XP");
               else if (ovInfo.dwMinorVersion == 2)
                    fprintf (fOutput, "Win 2003");
               else
                    fprintf (fOutput, "Unknown");
          } else
               fprintf (fOutput, "Unknown");

          if (!bCompact)
               fprintf (fOutput, "\n");
     } else
          ALARM (TA_SYS_ERR, "GetVersionEx");

     if (bCompact)
          fprintf (fOutput, "|");

     GetSystemTime (&stTime);
     fprintf (fOutput, (bCompact ? "%.2u/%.2u/%u %.2u:%.2u:%.2u.%.3u|" :
                        "UTC time : %.2u/%.2u/%u %.2u:%.2u:%.2u.%.3u\n"),
              stTime.wDay,
              stTime.wMonth,
              stTime.wYear,
              stTime.wHour,
              stTime.wMinute, stTime.wSecond, stTime.wMilliseconds);

     GetLocalTime (&stTime);
     fprintf (fOutput, (bCompact ? "%.2u/%.2u/%u %.2u:%.2u:%.2u.%.3u|" :
                        "Local time : %.2u/%.2u/%u %.2u:%.2u:%.2u.%.3u\n"),
              stTime.wDay,
              stTime.wMonth,
              stTime.wYear,
              stTime.wHour,
              stTime.wMinute, stTime.wSecond, stTime.wMilliseconds);

     fprintf (fOutput, (bCompact ? "\"%s\"|" : "Output locale: \"%s\"\n\n"),
	      dt_getlocale());

}
#else /* _WIN32 */
void
DumpSystemInfo (FILE * fOutput, const char *pContentName,
                const char *pMoreHeader, int bCompact)
{
     struct utsname infos;
     time_t chrono;
     struct tm *temps;

     if (bCompact) {
          fprintf (fOutput, "Content|Host Name|OS|UTC Time|Local Time|Output Locale|");
          fprintf (fOutput, "%s\n", pMoreHeader);
          fprintf (fOutput, "%s|", pContentName);
     }
     if (uname (&infos) == -1) {
          ALARM (TA_SYS_ERR, "uname");
          if (bCompact)
               fprintf (fOutput, "||");
     } else
          fprintf (fOutput, (bCompact ? "%s|%s %s %s|" :
                             "Hostname : %s\nOperating system : %s %s %s\n"),
                   infos.nodename, infos.sysname, infos.release,
                   infos.machine);

     chrono = time (NULL);
/* Not reentrant ! */
     temps = gmtime (&chrono);
     fprintf (fOutput, (bCompact ? "%.2u/%.2u/%u %.2u:%.2u:%.2u|" :
                        "UTC time : %.2u/%.2u/%u %.2u:%.2u:%.2u\n"),
              temps->tm_mday,
              temps->tm_mon + 1,
              temps->tm_year + 1900,
              temps->tm_hour, temps->tm_min, temps->tm_sec);

/* Not reentrant ! */
     temps = localtime (&chrono);
     fprintf (fOutput, (bCompact ? "%.2u/%.2u/%u %.2u:%.2u:%.2u|" :
                        "Local time : %.2u/%.2u/%u %.2u:%.2u:%.2u\n"),
              temps->tm_mday,
              temps->tm_mon + 1,
              temps->tm_year + 1900,
              temps->tm_hour, temps->tm_min, temps->tm_sec);

     fprintf (fOutput, (bCompact ? "\"%s\"|" : "Output locale: \"%s\"\n\n"),
	      dt_getlocale());

}
#endif /* _WIN32 */
